#!/usr/bin/env python

##################################################
# MODULE IMPORT
##################################################
# STANDARD MODULES
import os
import sys
import string
import logging
import numpy as np
import fnmatch
import shutil
import errno

# ASTRO MODULES
from astropy.io import fits
from astropy.io import ascii
from astropy.wcs import WCS
from astropy.nddata import Cutout2D
from astropy.stats import sigma_clipped_stats
from astropy import units as u
from astropy.coordinates import SkyCoord
from regions import CircleAnnulusSkyRegion, CircleAnnulusPixelRegion
import montage_wrapper as montage

# GRAPHICS MODULES
# import matplotlib.pyplot as plt


logger = logging.getLogger(__name__)


###########################
# CLASS DEFINITIONS
###########################
class Utils(object):
    """ Class collecting utility methods

                    Attributes:
                            None
    """

    def __init__(self):
        """ Return a Utils object """

    @classmethod
    def getBaseFile(cls, filename):
        """ Get basefilename without extension """
        filename_base = os.path.basename(filename)
        return filename_base

    @classmethod
    def getBaseFileNoExt(cls, filename):
        """ Get basefilename without extension """
        filename_base = os.path.basename(filename)
        filename_base_noext = os.path.splitext(filename_base)[0]
        return filename_base_noext

    @classmethod
    def mkdir(cls, path, delete_if_exists=False):
        """ Create a directory """
        try:
          if delete_if_exists and os.path.isdir(path):
            shutil.rmtree(path)
          os.makedirs(path)
        except OSError as exc:
          if exc.errno != errno.EEXIST:
            logger.error('Failed to create directory ' + path + '!')
            return -1

        return 0

    @classmethod
    def has_patterns_in_string(cls, s, patterns):
        """ Return true if patterns are found in string """
        if not patterns:
            return False

        found = False
        for pattern in patterns:
            found = pattern in s
            if found:
                break

        return found

    @classmethod
    def find_file(cls, rootPath, searched_file):
        """ Find file recursively from root directory and return full path """
        matches = []
        for root, dirnames, filenames in os.walk(rootPath):
            for filename in fnmatch.filter(filenames, searched_file):
                matches.append(os.path.join(root, filename))
        return matches

    @classmethod
    def find_and_copy_file(cls, rootPath, searched_file, destDir):
        """ Find file recursively from root dir and move to destination dir """

        matched_files = Utils.find_file(rootPath, searched_file)
        if not matched_files:
            return -1

        for filename in matched_files:
            shutil.copy(filename, destDir)

        return 0

    @classmethod
    def write_ascii(cls, data, filename, header=''):
        """ Write data to ascii file """

        # - Skip if data is empty
        if data.size <= 0:
            # cls._logger.warn("Empty data given, no file will be written!")
            logger.warning("Empty data given, no file will be written!")
            return

        # - Open file and write header
        fout = open(filename, 'wt')
        if header:
            fout.write(header)
            fout.write('\n')
            fout.flush()

        # - Write data to file
        nrows = data.shape[0]
        ncols = data.shape[1]
        for i in range(nrows):
            fields = '  '.join(map(str, data[i, :]))
            fout.write(fields)
            fout.write('\n')
            fout.flush()

        fout.close()

    @classmethod
    def read_ascii(cls, filename, skip_patterns=[]):
        """ Read an ascii file line by line """

        try:
            f = open(filename, 'r')
        except IOError:
            errmsg = 'Could not read file: ' + filename
            # cls._logger.error(errmsg)
            logger.error(errmsg)
            raise IOError(errmsg)

        fields = []
        for line in f:
            line = line.strip()
            line_fields = line.split()
            if not line_fields:
                continue

            # Skip pattern
            skipline = cls.has_patterns_in_string(
                line_fields[0], skip_patterns)
            if skipline:
                continue

            fields.append(line_fields)

        f.close()

        return fields

    @classmethod
    def read_ascii_table(cls, filename, row_start=0, delimiter='|', format=''):
        """ Read an ascii table file line by line """

        try:
            if format:
                table = ascii.read(filename, format=format)
            else:
                table = ascii.read(
                    filename, data_start=row_start, delimiter=delimiter)
        except Exception as ex:
            logger.error("Failed to read table!")
            return None

        return table

    @classmethod
    def write_fits(cls, data, filename, header=None):
        """ Read data to FITS image """
        if header:
            hdu = fits.PrimaryHDU(data, header)
        else:
            hdu = fits.PrimaryHDU(data)

        hdul = fits.HDUList([hdu])
        hdul.writeto(filename, overwrite=True)

    @classmethod
    def read_fits(cls, filename):
        """ Read FITS image and return data """

        # - Open file
        try:
            hdu = fits.open(filename, memmap=False)
        except Exception as ex:
            errmsg = 'Cannot read image file: ' + filename
            # cls._logger.error(errmsg)
            logger.error(errmsg)
            raise IOError(errmsg)

        # - Read data
        data = hdu[0].data
        data_size = np.shape(data)
        nchan = len(data.shape)
        if nchan == 4:
            output_data = data[0, 0, :, :]
        elif nchan == 2:
            output_data = data
        else:
            errmsg = 'Invalid/unsupported number of channels found in file ' + \
                filename + ' (nchan=' + str(nchan) + ')!'
            # cls._logger.error(errmsg)
            logger.error(errmsg)
            hdu.close()
            raise IOError(errmsg)

        # - Read metadata
        header = hdu[0].header

        # - Close file
        hdu.close()

        return output_data, header

    @classmethod
    def crop_img(cls, data, x0, y0, dx, dy):
        """ Extract sub image of size (dx,dy) around pixel (x0,y0) """

        # - Extract crop data
        xmin = int(x0-dx/2)
        xmax = int(x0+dx/2)
        ymin = int(y0-dy/2)
        ymax = int(y0+dy/2)
        crop_data = data[ymin:ymax+1, xmin:xmax+1]

        # - Replace NAN with zeros and inf with large numbers
        np.nan_to_num(crop_data, False)

        return crop_data

    @classmethod
    def getBeamArea(cls, bmaj, bmin):
        """ Compute beam area """
        beamArea = np.pi * bmaj * bmin/(4.*np.log(2))
        return beamArea

    @classmethod
    def getNVSSSurveyBeamArea(cls):
        """ Returns NVSS survey beam area """
        bmaj = 45.  # arcsec
        bmin = 45.  # arcsec
        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getVLASSSurveyBeamArea(cls):
        """ Returns VLASS survey beam area """
        # Very approximated (no pa info), taken from Lacy arXiv:1907.01981v3 (2019)
        bmaj = 3.  # arcsec
        bmin = 2.  # arcsec (assumed 1.5 smaller than bmaj)
        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getWiseSurveyBeamArea(cls, band):
        """ Returns Wise survey beam area """
        if band == 'wise_3_4':     # http://wise2.ipac.caltech.edu/docs/release/allsky/  (Wrigth+10)
            bmaj = 6.1  # arcsec
            bmin = 6.1  # arcsec
        elif band == 'wise_4_6':
            bmaj = 6.4  # arcsec
            bmin = 6.4  # arcsec
        elif band == 'wise_12':
            bmaj = 6.5  # arcsec
            bmin = 6.5  # arcsec
        elif band == 'wise_22':
            bmaj = 12  # arcsec
            bmin = 12  # arcsec
        else:
            logger.error("Invalid/unknown band argument given (" + band + ")!")
            return 0

        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getIRACSurveyBeamArea(cls, band):
        """ Returns IRAC survey beam area """
        if band == 'irac_3_6':
            bmaj = 1.7  # arcsec
            bmin = 1.7  # arcsec
        elif band == 'irac_4_5':
            bmaj = 1.7  # arcsec
            bmin = 1.7  # arcsec
        elif band == 'irac_5_8':
            bmaj = 1.7  # arcsec
            bmin = 1.7  # arcsec
        elif band == 'irac_8':
            bmaj = 1.9  # arcsec
            bmin = 1.9  # arcsec
        else:
            logger.error("Invalid/unknown band argument given (" + band + ")!")
            return 0

        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getMIPSSurveyBeamArea(cls):
        """ Returns MIPS survey beam area """

        # arcsec    Rieke+04 (https://iopscience.iop.org/article/10.1086/422717/pdf)
        bmaj = 6
        bmin = 6  # arcsec
        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getHiGalSurveyBeamArea(cls, band):
        """ Returns HiGal survey beam area """
        if band == 'higal_70':
            bmaj = 6.7  # arcsec  https://hi-gal.iaps.inaf.it/ ; Bufano+18
            bmin = 6.7  # arcsec
        elif band == 'higal_160':
            bmaj = 11.  # arcsec
            bmin = 11.  # arcsec
        elif band == 'higal_250':
            bmaj = 18.  # arcsec
            bmin = 18.  # arcsec
        elif band == 'higal_350':
            bmaj = 25.  # arcsec
            bmin = 25.  # arcsec
        elif band == 'higal_500':
            bmaj = 37.  # arcsec
            bmin = 37.  # arcsec
        else:
            logger.error("Invalid/unknown band argument given (" + band + ")!")
            return 0

        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getATLASGALSurveyBeamArea(cls):
        """ Returns ATLASGAL survey beam area """

        # arcsec (from http://www.apex-telescope.org/bolometer/laboca/technical/)
        bmaj = 18.6
        # arcsec (from http://www.apex-telescope.org/bolometer/laboca/technical/)
        bmin = 18.6
        # bmaj= 19.2 # arcsec (from ATLASGAL papers)
        # bmin= 19.2 # arcsec (from ATLASGAL papers)
        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getATLASGALPlanckSurveyBeamArea(cls):
        """ Returns ATLASGAL+Planck survey beam area """

        # arcsec (from http://atlasgal.mpifr-bonn.mpg.de/cgi-bin/ATLASGAL_DATASETS.cgi)
        bmaj = 21
        bmin = 21  # arcsec
        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getMSXSurveyBeamArea(cls):
        """ Returns MSX survey beam area """

        bmaj = 20  # arcsec	http://irsa.ipac.caltech.edu/applications/MSX/MSX/imageDescriptions.htm
        bmin = 20  # arcsec
        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getFIRSTSurveyBeamArea(cls, ra, dec):
        """ Returns FIRST survey beam area """

        if not ra or not dec:
            bmaj = 5.4
            bmin = 5.4

        if dec > 4.55583333:
            bmaj = 5.4
            bmin = 5.4
        else:
            if dec < -2.50694444 and (ra < 45 or ra > 315):
                bmaj = 6.8
                bmin = 5.4
            else:
                bmaj = 6.4
                bmin = 5.4

        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getMGPSSurveyBeamArea(cls, dec):
        """ Returns MGPS survey beam area """

        if not dec or dec == 0:
            bmaj = 43
        else:
            bmaj = 43./np.sin(np.deg2rad(dec))  # arcsec
        bmin = 43.  # arcsec

        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getSGPSSurveyBeamArea(cls):
        """ Returns SGPS survey beam area """

        # - Taken from SGPS images
        #   NB: In https://iopscience.iop.org/article/10.1086/430114/pdf slightly different values (depending on l/b) are given
        bmin = 100.  # arcsec
        bmaj = 100.  # arcsec
        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getVGPSSurveyBeamArea(cls):
        """ Returns VGPS survey beam area """

        # - Taken from https://iopscience.iop.org/article/10.1086/505940/pdf
        bmin = 45.  # arcsec
        bmaj = 45.  # arcsec
        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getScorpioATCASurveyBeamArea(cls, band):
        """ Returns Scorpio ATCA survey beam area """

        if band == 'scorpio_atca_2_1':
            bmaj = 9.8  # arcsec
            bmin = 5.8  # arcsec
        # elif band=='scorpio_atca_5_0':
        #	bmaj= 12.103 # arcsec
        #	bmin= 12.103 # arcsec

        else:
            logger.error("Invalid/unknown band argument given (" + band + ")!")
            return 0

        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getScorpioASKAP15B1SurveyBeamArea(cls):
        """ Returns Scorpio ASKAP 15 B1 survey beam area """
        bmaj = 24  # arcsec
        bmin = 21  # arcsec
        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getScorpioASKAP36B123SurveyBeamArea(cls):
        """ Returns Scorpio ASKAP 36 B123 survey beam area """
        bmaj = 9.403  # arcsec
        bmin = 7.713  # arcsec
        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getScorpioASKAP36B123SubChanSurveyBeamArea(cls):
        """ Returns Scorpio ASKAP 36 B123 ch1 survey beam area """
        bmaj = 14.0  # arcsec
        bmin = 12.0  # arcsec
        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getTHORSurveyBeamArea(cls):
        """ Returns THOR survey beam area """
        # Beam varying from 18.1 x 11.1 to 12.0 x 11.6
        bmaj = 12  # arcsec
        bmin = 12  # arcsec
        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getMeerkatGPSSurveyBeamArea(cls):
        """ Returns Meerkat GPS survey beam area """
        bmaj = 8  # arcsec
        bmin = 8  # arcsec
        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getASKAPRACSSurveyBeamArea(cls):
        """ Returns ASKAP RACS survey beam area """
        bmaj = 25  # arcsec
        bmin = 25  # arcsec
        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getMAGPIS21cmSurveyBeamArea(cls):
        """ Returns MAGPIS new 21cm survey beam area """
        bmaj = 6.2  # arcsec
        bmin = 5.4  # arcsec
        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getCORNISHSurveyBeamArea(cls):
        """ Returns CORNISH 5 GHz survey beam area """
        bmaj = 1.5  # arcsec
        bmin = 1.5  # arcsec
        bmaj_deg = bmaj/3600.
        bmin_deg = bmin/3600.
        beamArea = Utils.getBeamArea(bmaj_deg, bmin_deg)
        return beamArea

    @classmethod
    def getSurveyBeamArea(cls, survey, ra=None, dec=None):
        """ Return beam area of given survey """

        beamArea = 0
        if survey == 'nvss':
            beamArea = Utils.getNVSSSurveyBeamArea()
        elif survey == 'first':
            beamArea = Utils.getFIRSTSurveyBeamArea(ra, dec)
        elif survey == 'mgps':
            beamArea = Utils.getMGPSSurveyBeamArea(dec)
        elif survey == 'sgps':
            beamArea = Utils.getSGPSSurveyBeamArea()
        elif survey == 'vgps':
            beamArea = Utils.getVGPSSurveyBeamArea()
        elif survey == 'vlass':
            beamArea = Utils.getVLASSSurveyBeamArea()
        elif survey == 'wise_3_4':
            beamArea = Utils.getWiseSurveyBeamArea(survey)
        elif survey == 'wise_4_6':
            beamArea = Utils.getWiseSurveyBeamArea(survey)
        elif survey == 'wise_12':
            beamArea = Utils.getWiseSurveyBeamArea(survey)
        elif survey == 'wise_22':
            beamArea = Utils.getWiseSurveyBeamArea(survey)
        elif survey == 'irac_3_6':
            beamArea = Utils.getIRACSurveyBeamArea(survey)
        elif survey == 'irac_4_5':
            beamArea = Utils.getIRACSurveyBeamArea(survey)
        elif survey == 'irac_5_8':
            beamArea = Utils.getIRACSurveyBeamArea(survey)
        elif survey == 'irac_8':
            beamArea = Utils.getIRACSurveyBeamArea(survey)
        elif survey == 'mips_24':
            beamArea = Utils.getMIPSSurveyBeamArea()
        elif survey == 'higal_70':
            beamArea = Utils.getHiGalSurveyBeamArea(survey)
        elif survey == 'higal_160':
            beamArea = Utils.getHiGalSurveyBeamArea(survey)
        elif survey == 'higal_250':
            beamArea = Utils.getHiGalSurveyBeamArea(survey)
        elif survey == 'higal_350':
            beamArea = Utils.getHiGalSurveyBeamArea(survey)
        elif survey == 'higal_500':
            beamArea = Utils.getHiGalSurveyBeamArea(survey)
        elif survey == 'atlasgal':
            beamArea = Utils.getATLASGALSurveyBeamArea()
        elif survey == 'atlasgal_planck':
            beamArea = Utils.getATLASGALPlanckSurveyBeamArea()
        elif survey == 'msx_8_3':
            beamArea = Utils.getMSXSurveyBeamArea()
        elif survey == 'msx_12_1':
            beamArea = Utils.getMSXSurveyBeamArea()
        elif survey == 'msx_14_7':
            beamArea = Utils.getMSXSurveyBeamArea()
        elif survey == 'msx_21_3':
            beamArea = Utils.getMSXSurveyBeamArea()
        elif survey == 'scorpio_atca_2_1':
            beamArea = Utils.getScorpioATCASurveyBeamArea(survey)
        elif survey == 'scorpio_askap15_b1':
            beamArea = Utils.getScorpioASKAP15B1SurveyBeamArea()
        elif survey == 'scorpio_askap36_b123':
            beamArea = Utils.getScorpioASKAP36B123SurveyBeamArea()
        elif survey == 'scorpio_askap36_b123_ch1':
            beamArea = Utils.getScorpioASKAP36B123SubChanSurveyBeamArea()
        elif survey == 'scorpio_askap36_b123_ch2':
            beamArea = Utils.getScorpioASKAP36B123SubChanSurveyBeamArea()
        elif survey == 'scorpio_askap36_b123_ch3':
            beamArea = Utils.getScorpioASKAP36B123SubChanSurveyBeamArea()
        elif survey == 'scorpio_askap36_b123_ch4':
            beamArea = Utils.getScorpioASKAP36B123SubChanSurveyBeamArea()
        elif survey == 'scorpio_askap36_b123_ch5':
            beamArea = Utils.getScorpioASKAP36B123SubChanSurveyBeamArea()
        elif survey == 'thor':
            beamArea = Utils.getTHORSurveyBeamArea()
        elif survey == 'meerkat_gps':
            beamArea = Utils.getMeerkatGPSSurveyBeamArea()
        elif survey == 'meerkat_gps_ch1':
            beamArea = Utils.getMeerkatGPSSurveyBeamArea()
        elif survey == 'meerkat_gps_ch2':
            beamArea = Utils.getMeerkatGPSSurveyBeamArea()
        elif survey == 'meerkat_gps_ch3':
            beamArea = Utils.getMeerkatGPSSurveyBeamArea()
        elif survey == 'meerkat_gps_ch4':
            beamArea = Utils.getMeerkatGPSSurveyBeamArea()
        elif survey == 'meerkat_gps_ch5':
            beamArea = Utils.getMeerkatGPSSurveyBeamArea()
        elif survey == 'meerkat_gps_ch6':
            beamArea = Utils.getMeerkatGPSSurveyBeamArea()
        elif survey == 'meerkat_gps_ch7':
            beamArea = Utils.getMeerkatGPSSurveyBeamArea()
        elif survey == 'meerkat_gps_ch8':
            beamArea = Utils.getMeerkatGPSSurveyBeamArea()
        elif survey == 'meerkat_gps_ch9':
            beamArea = Utils.getMeerkatGPSSurveyBeamArea()
        elif survey == 'meerkat_gps_ch10':
            beamArea = Utils.getMeerkatGPSSurveyBeamArea()
        elif survey == 'meerkat_gps_ch11':
            beamArea = Utils.getMeerkatGPSSurveyBeamArea()
        elif survey == 'meerkat_gps_ch12':
            beamArea = Utils.getMeerkatGPSSurveyBeamArea()
        elif survey == 'meerkat_gps_ch13':
            beamArea = Utils.getMeerkatGPSSurveyBeamArea()
        elif survey == 'meerkat_gps_ch14':
            beamArea = Utils.getMeerkatGPSSurveyBeamArea()
        elif survey == 'askap_racs':
            beamArea = Utils.getASKAPRACSSurveyBeamArea()
        elif survey == 'magpis_21cm':
            beamArea = Utils.getMAGPIS21cmSurveyBeamArea()
        elif survey == 'cornish':
            beamArea = Utils.getCORNISHSurveyBeamArea()
        else:
            logger.error("Unknown survey (" + survey + "), returning area=0!")
            beamArea = 0

        return beamArea

    @classmethod
    def getJyBeamToPixel(cls, beamArea, dx, dy):
        """ Compute conversion factor from Jy/beam to Jy/pixel """

        pixArea = np.abs(dx*dy)
        toJyPerPix = pixArea/beamArea

        return toJyPerPix

    @classmethod
    def getJyBeamToPixel2(cls, bmaj, bmin, dx, dy):
        """ Compute conversion factor from Jy/beam to Jy/pixel """
        beamArea = Utils.getBeamArea(bmaj, bmin)
        return Utils.getJyBeamToPixel(beamArea, dx, dy)

    @classmethod
    def hasBeamInfo(cls, header):
        """ Check if header has beam information """
        hasBmaj = ('BMAJ' in header)
        hasBmin = ('BMIN' in header)
        hasBeamInfo = (hasBmaj and hasBmin)
        if not hasBeamInfo:
            return False

        hasBeamVal = (header['BMAJ'] and header['BMIN'])
        if not hasBeamVal:
            return False

        return True

    @classmethod
    def fixImgAxisAndUnits(cls, filename, outfile):
        """ Fix image axis issues and convert units """

        # - Open input file
        try:
            hdu = fits.open(filename, memmap=False)
        except Exception as ex:
            logger.error('Cannot read image file: ' + filename)
            return -1

        # - Read data & header
        header = hdu[0].header
        data = hdu[0].data
        data_size = np.shape(data)
        nchan = len(data.shape)
        if nchan == 4:
            output_data = data[0, 0, :, :]
            output_header = header
            output_header['NAXIS'] = 2

        elif nchan == 2:
            output_data = data
            output_header = header

        else:
            errmsg = 'Invalid/unsupported number of channels found in file ' + \
                filename + ' (nchan=' + str(nchan) + ')!'
            hdu.close()
            logger.error(errmsg)
            return -1

        logger.info("Deleting NAXIS3/NAXIS4 from header...")
        if 'NAXIS3' in output_header:
            del output_header['NAXIS3']
        if 'NAXIS4' in output_header:
            del output_header['NAXIS4']

        if 'CTYPE3' in output_header:
            del output_header['CTYPE3']
        if 'CRVAL3' in output_header:
            del output_header['CRVAL3']
        if 'CDELT3' in output_header:
            del output_header['CDELT3']
        if 'CRPIX3' in output_header:
            del output_header['CRPIX3']
        if 'CROTA3' in output_header:
            del output_header['CROTA3']

        if 'CTYPE4' in output_header:
            del output_header['CTYPE4']
        if 'CRVAL4' in output_header:
            del output_header['CRVAL4']
        if 'CDELT4' in output_header:
            del output_header['CDELT4']
        if 'CRPIX4' in output_header:
            del output_header['CRPIX4']
        if 'CROTA4' in output_header:
            del output_header['CROTA4']

        # - Close input file
        hdu.close()

        # - Scale flux?
        bscale = 1
        bzero = 0
        if 'BZERO' in output_header:
            bzero = output_header['BZERO']
        if 'BSCALE' in output_header:
            bscale = output_header['BSCALE']

        if bzero != 0 or bscale != 1:
            logger.info("Scaling image " + filename + " flux by bscale=" +
                        str(bscale) + ", bzero=" + str(bzero) + " ...")
            output_data_scaled = bzero + bscale*output_data
            output_data = output_data_scaled
            output_header['BSCALE'] = 1
            output_header['BZERO'] = 0

        # - Add missing BMAJ/BMIN to header?
        # ...

        # - Convert data to float 32
        output_data = output_data.astype(np.float32)

        # - Write "fixed" fits
        Utils.write_fits(output_data, outfile, output_header)

        return 0

    @classmethod
    def fromBrightnessTempToJyBeam(cls, nu_GHz, bmin_arcsec, bmaj_arcsec):
        """ Convert flux density from brightness temperature to Jy/beam """

        # - Taken from https://science.nrao.edu/facilities/vla/proposing/TBconv
        # T= 1.222e+3*I/(nu_GHz*nu_GHz*bmin_arcsec*bmaj_arcsec)
        I_mJyBeam = nu_GHz*nu_GHz*bmin_arcsec*bmaj_arcsec/1.222e+3
        I_JyBeam = I_mJyBeam*1.e-3
        return I_JyBeam

    @classmethod
    def convertImgToJyPixel(cls, filename, outfile, survey='',default_bunit=''):
        """ Convert image units from original to Jy/pixel """

        # - Read fits image
        data, header = Utils.read_fits(filename)
        wcs = WCS(header)

        # - Read WCS axis name
        wcs_axis_types= wcs.axis_type_names
        wcs_axis_units= wcs.world_axis_units
        n_wcs_axis= wcs.naxis
        is_galactic= (wcs_axis_types[0]=='GLON') and (wcs_axis_types[1]=='GLAT')

        # - Check header keywords
        if 'BUNIT' not in header:
            if not default_bunit:
                logger.error("No available BUNIT, cannot compute conversion factor!")
                return -1
            else:
                header['BUNIT'] = default_bunit
                logger.warning("No BUNIT keyword present in file " + filename + ", using value read from metadata table!")
                
        if 'CDELT1' not in header:
            logger.error("No CDELT1 keyword present in file " + filename + ", cannot compute conversion factor!")
            return -1

        if 'CDELT2' not in header:
            logger.error("No CDELT2 keyword present in file " + filename + ", cannot compute conversion factor!")
            return -1

        if 'CRPIX1' not in header:
            logger.error("No CRPIX1 keyword present in file " + filename + ", cannot compute conversion factor!")
            return -1

        if 'CRPIX2' not in header:
            logger.error("No CRPIX2 keyword present in file " + filename + ", cannot compute conversion factor!")
            return -1

        units = header['BUNIT'].strip()
        dx = abs(header['CDELT1'])  # in deg
        dy = abs(header['CDELT2'])  # in deg

        # ==================
        # - Convert data
        # ==================
        convFactor = 1

        # - Jy/beam units (e.g. radio maps, apex)
        if units == 'JY/BEAM' or units == 'Jy/beam':
            xc = header['CRPIX1']
            yc = header['CRPIX2']
            
            hasBeamInfo = Utils.hasBeamInfo(header)
            if hasBeamInfo:
                bmaj = header['BMAJ']  # in deg
                bmin = header['BMIN']  # in deg
                convFactor = Utils.getJyBeamToPixel2(bmaj, bmin, dx, dy)
            else:
                logger.warning("No BMAJ/BMIN keyword present in file " + filename + ", trying to retrieve from survey name...")

                if is_galactic:
                  try:
                    if n_wcs_axis==2:
                      l, b = wcs.all_pix2world(xc, yc, 0)
                    elif n_wcs_axis==3:
                      l, b, _ = wcs.all_pix2world(xc, yc, 0, 0)
                    elif n_wcs_axis==4:
                      l, b, _, _ = wcs.all_pix2world(xc, yc, 0, 0, 0)
                    else:
                      l, b = (0, 0)
                      logger.warning("Cannot compute glon/glat from WCS header as naxis is not 2, 3 or 4, assuming (0,0)")

                    if l!=0 and b!=0:
                      coord_gal = SkyCoord(l=l, b=b, unit='deg', frame="galactic")
                      coord_j2000= coord_gal.transform_to("fk5")
                      ra= coord_j2000.ra.value
                      dec= coord_j2000.dec.value

                  except Exception:
                    ra, dec = (0, 0)
                    logger.warning("Cannot compute RA, Dec from WCS header, assuming (0,0)")
                else:
                  try:
                    if n_wcs_axis==2:
                      ra, dec = wcs.all_pix2world(xc, yc, 0, ra_dec_order=True)
                    elif n_wcs_axis==3:
                      ra, dec, _ = wcs.all_pix2world(xc, yc, 0, 0, ra_dec_order=True)
                    elif n_wcs_axis==4:
                      ra, dec, _, _ = wcs.all_pix2world(xc, yc, 0, 0, 0, ra_dec_order=True)
                    else:
                      ra, dec = (0, 0)
                      logger.warning("Cannot compute RA, Dec from WCS header as naxis is not 2, 3 or 4, assuming (0,0)")
                  except Exception:
                    ra, dec = (0, 0)
                    logger.warning("Cannot compute RA, Dec from WCS header, assuming (0,0)")

                beamArea = Utils.getSurveyBeamArea(survey, ra, dec)
                if beamArea > 0:
                    convFactor = Utils.getJyBeamToPixel(beamArea, dx, dy)
                else:
                    logger.error("No BMAJ keyword present in file " +
                                 filename + ", cannot compute conversion factor!")
                    return -1

        # - DN UNITS (e.g WISE MAPS)
        elif units == 'DN':
            if survey == 'wise_3_4':
                convFactor = 1.9350E-06
            elif survey == 'wise_4_6':
                convFactor = 2.7048E-06
            elif survey == 'wise_12':
                convFactor = 1.8326e-06
            elif survey == 'wise_22':
                convFactor = 5.2269E-05
            else:
                logger.error(
                    "Invalid or unknown survey (" + survey + ") given!")
                return -1

        # - MJy/sr units (e.g. HERSCHEL/SPITZER maps)
        elif units.startswith('MJy/sr'):
            # we need dx and dy in ''
            convFactor = 1.e+6*(3600*dx)*(3600*dy)/(206265.*206265.)

        # - W/m^2-sr units (e.g. MSX maps)
        elif units == 'W/m^2-sr':
            # we need dx and dy in ''
            convFactor_fromJyToSr = (3600*dx)*(3600*dy)/(206265.*206265.)
            if survey == 'msx_8_3':
                convFactor = 7.133e+12*convFactor_fromJyToSr
            elif survey == 'msx_12_1':
                convFactor = 2.863e+13*convFactor_fromJyToSr
            elif survey == 'msx_14_7':
                convFactor = 3.216e+13*convFactor_fromJyToSr
            elif survey == 'msx_21_3':
                convFactor = 2.476e+13*convFactor_fromJyToSr
            else:
                logger.error(
                    "Invalid or unknown survey (" + survey + ") given!")
                return -1

        # - T units (e.g. VGPS)
        elif units == 'K':
            if survey == 'vgps':
                nu = 1.4  # GHz
                bmaj = 45  # arcsec
                bmin = 45  # arcsec
                convFactor = Utils.fromBrightnessTempToJyBeam(nu, bmin, bmaj)
            else:
                logger.error(
                    "Invalid or unknown survey (" + survey + ") given!")
                return -1

        # - Jy/pixel units (e.g. simulated maps)
        elif units == 'Jy/pixel' or units == 'JY/PIXEL':
            convFactor = 1
        else:
            logger.error('Units ' + units + ' not recognized!')
            return -1

        # - Scale data
        data_conv = data*convFactor
        header_conv = header

        # - Edit header
        header_conv['BUNIT'] = 'Jy/pixel'

        # - Write converted fits to file
        Utils.write_fits(data_conv, outfile, header_conv)

        return 0

    @classmethod
    def cropImage(cls, filename, ra, dec, crop_mode, crop_size, outfile, source_size=-1, nanfill=True, nanfill_mode='imgmin', nanfill_val=0):
        """ Crop image around (ra,dec) by crop_size """

        # - Read fits image
        data, header = Utils.read_fits(filename)
        wcs = WCS(header)
        dx = abs(header['CDELT1'])  # in deg
        dy = abs(header['CDELT2'])  # in deg
        pix_size = max(dx, dy)  # in deg
        data_shape= data.shape 
        ndim= data.ndim
        
        if ndim!=2:
          if ndim==3:
            data= data[0, :, :]
          elif n_wcs_axis==4:
            data= data[0, 0, :, :]
          else:
            errmsg= "WCS naxis is " + str(n_wcs_axis) + ", cannot extract image data!" 
            logger.warning(errmsg)
            raise Exception(errmsg)

        # - Check mismatch between wcs naxis and data
        ndim= data.ndim
        n_wcs_axis= wcs.naxis
        if ndim!=n_wcs_axis:
          logger.info("Data and WCS axis differs, trying to remove 3rd & 4th axis header keywords ...") 

          header['NAXIS']= 2			
          if 'NAXIS3' in header:
            del header['NAXIS3']
          if 'NAXIS4' in header:
            del header['NAXIS4']
          if 'CTYPE3' in header:
            del header['CTYPE3']
          if 'CRVAL3' in header:
            del header['CRVAL3']
          if 'CDELT3' in header:
            del header['CDELT3']
          if 'CRPIX3' in header:
            del header['CRPIX3']
          if 'CROTA3' in header:
            del header['CROTA3']
          if 'CUNIT3' in header:
            del header['CUNIT3']
          if 'CTYPE4' in header:
            del header['CTYPE4']
          if 'CRVAL4' in header:
            del header['CRVAL4']
          if 'CDELT4' in header:
            del header['CDELT4']
          if 'CRPIX4' in header:
            del header['CRPIX4']
          if 'CROTA4' in header:
            del header['CROTA4']
          if 'CUNIT4' in header:
            del header['CUNIT4']
 
          if 'PC03_01' in header:
            del header['PC03_01']
          if 'PC04_01' in header:
            del header['PC04_01']
          if 'PC03_02' in header:
            del header['PC03_02']
          if 'PC04_02' in header:
            del header['PC04_02']
          if 'PC01_03' in header:
            del header['PC01_03']
          if 'PC02_03' in header:
            del header['PC02_03']
          if 'PC03_03' in header:
            del header['PC03_03']
          if 'PC04_03' in header:
            del header['PC04_03']
          if 'PC01_04' in header:
            del header['PC01_04']
          if 'PC02_04' in header:
            del header['PC02_04']
          if 'PC03_04' in header:
            del header['PC03_04']
          if 'PC04_04' in header:
            del header['PC04_04']

          if 'PC3_1' in header:
            del header['PC3_1']
          if 'PC4_1' in header:
            del header['PC4_1']
          if 'PC3_2' in header:
            del header['PC3_2']
          if 'PC4_2' in header:
            del header['PC4_2']
          if 'PC1_3' in header:
            del header['PC1_3']
          if 'PC2_3' in header:
            del header['PC2_3']
          if 'PC3_3' in header:
            del header['PC3_3']
          if 'PC4_3' in header:
            del header['PC4_3']
          if 'PC1_4' in header:
            del header['PC1_4']
          if 'PC2_4' in header:
            del header['PC2_4']
          if 'PC3_4' in header:
            del header['PC3_4']
          if 'PC4_4' in header:
            del header['PC4_4']

          logger.info("Recreating WCS after modyfying the header ...") 
          wcs = WCS(header)
          n_wcs_axis= wcs.naxis
          if ndim!=n_wcs_axis:
            errmsg= "Data and WCS axis (" + str(n_wcs_axis) + ") still differs, giving up ..."
            logger.error(errmsg) 
            print("FITS header")
            print(header)
            raise Exception(errmsg)

        # - Read WCS axis name
        wcs_axis_types= wcs.axis_type_names
        wcs_axis_units= wcs.world_axis_units
        n_wcs_axis= wcs.naxis
        is_galactic= (wcs_axis_types[0]=='GLON') and (wcs_axis_types[1]=='GLAT')

        # - Check if crop size is cutting part of the source
        if source_size != -1:
            source_size_pix = source_size/pix_size

        crop_size_pix = 0

        if crop_mode == 'pixel':
            crop_size_pix = crop_size
            if source_size_pix >= crop_size_pix:
                logger.warning("Requested crop size (%d) smaller than source size (size=%d arcsec, %d pix), won't write cropped fits file!" % (crop_size_pix, source_size*3600, source_size_pix))
                return -1
        elif crop_mode == 'factor':
            crop_size_pix = 2 * crop_size * source_size_pix # twice the source radius*factor

        logging.info('Cropping image to {0}x{0} px'.format(crop_size_pix, crop_size_pix))

        # - Find pixel coordinates corresponding to ra,dec
        if is_galactic:
          sc = SkyCoord(ra=ra, dec=dec, unit='deg', frame="fk5")
          sc_gal= sc.transform_to("galactic")
          l= sc_gal.l.value
          b= sc_gal.b.value

          if n_wcs_axis==2:
            x0, y0 = wcs.all_world2pix(l, b, 0)
          elif n_wcs_axis==3:
            x0, y0, _ = wcs.all_world2pix(l, b, 0, 0)
          elif n_wcs_axis==4:
            x0, y0, _, _ = wcs.all_world2pix(l, b, 0, 0, 0)
          else:
            errmsg= "WCS naxis is not 2, 3 or 4, cannot compute pix coordinates!" 
            logger.warning(errmsg)
            raise Exception(errmsg)

        else:
          if n_wcs_axis==2:
            x0, y0 = wcs.all_world2pix(ra, dec, 0, ra_dec_order=True)
          elif n_wcs_axis==3:
            x0, y0, _ = wcs.all_world2pix(ra, dec, 0, 0, ra_dec_order=True)
          elif n_wcs_axis==4:
            x0, y0, _, _ = wcs.all_world2pix(ra, dec, 0, 0, 0, ra_dec_order=True) 
          else:
            errmsg= "WCS naxis is not 2, 3 or 4, cannot compute pix coordinates!" 
            logger.warning(errmsg)
            raise Exception(errmsg)
        
        # - Extract cutout. With option 'partial' when cutout size is larger than image size the cutout will be filled with nan (or specified value)
        try:
            cutout = Cutout2D(data, (x0, y0), (crop_size_pix,crop_size_pix), mode='partial', wcs=wcs)
        except Exception as e:
            logger.error("Failed to create cutout (err=%s)!" % str(e))
            return -1

        output_data = cutout.data

        # - Fill NAN?
        if nanfill:
            if nanfill_mode == 'imgmin':
                img_min = np.nanmin(output_data)
                output_data[np.isnan(output_data)] = img_min
            else:
                output_data[np.isnan(output_data)] = nanfill_val

        # - Create new header copying old one but overriding NAXIS keywords
        #   NB: Not working, need to change also CRPIX, use cutout.wcs.to_header()
        output_header= header
        output_header.update(cutout.wcs.to_header())
        #output_header['NAXIS']= 2
        #output_header['NAXIS1']= output_data.shape[1]
        #output_header['NAXIS2']= output_data.shape[0]

        # - Write reshaped image fits
        Utils.write_fits(output_data, outfile, output_header)
        #Utils.write_fits(output_data, outfile)


        return 0

    @classmethod
    def estimateBkgFromAnnulus(cls, filename, ra, dec, R1, R2, method='sigmaclip', max_nan_thr=0.1):
        """ Estimate bkg from annulus around given sky position """

        # - Read fits image
        data, header = Utils.read_fits(filename)
        wcs = WCS(header)

        # - Define sky annulus region
        center = SkyCoord(ra*u.deg, dec*u.deg, frame='fk5')
        annulus_sky = CircleAnnulusSkyRegion(
            center=center,
            inner_radius=R1*u.arcsec,
            outer_radius=R2*u.arcsec
        )

        # - Convert sky annulus to pixel annulus
        annulus_pix = annulus_sky.to_pixel(wcs)

        # - Get mask corresponding to annulus and array of pixel values in the mask
        mask = annulus_pix.to_mask()
        mask_data = mask.to_image(shape=data.shape)
        aperture_pixel_data = data[mask_data == 1]

        # - Integrity check before computing stats
        npixels = aperture_pixel_data.size
        if npixels == 0:
            logger.warning("Pixel mask is empty, returning zero!")
            return 0

        nan_counts = np.count_nonzero(np.isnan(aperture_pixel_data))
        nan_fract = float(nan_counts)/float(npixels)
        if nan_fract > max_nan_thr:
            logger.warning("Fraction of nan pixels in mask (%s) exceeding max threshold (%s), returning zero!" % (
                str(nan_fract), str(max_nan_thr)))
            return 0

        # - Compute median or 3-sigma clipped value as bkg estimator
        bkg = 0
        nsigma = 3
        if method == 'median':
            bkg = np.median(aperture_pixel_data)
        elif method == 'sigmaclip':
            mean, median, rms = sigma_clipped_stats(
                aperture_pixel_data, sigma=nsigma)
            bkg = median
        else:
            logger.error(
                "Invalid bkg estimation method (%s) given, returning zero!" % (method))
            return -1

        return bkg

    @classmethod
    def makeMosaic(cls, input_tbl, output, combine="mean", background_match=False, bitpix=-32, exact=False):
        """ Create a mosaic from input images """

        # - Get output file base dir
        outdir = os.path.dirname(output)

        # - Get current directory and create work dir
        currdir = os.path.dirname(os.path.realpath(input_tbl))
        input_tbl_base = Utils.getBaseFileNoExt(input_tbl)
        workdir = 'mosaic_' + input_tbl_base
        dir_path = os.path.join(currdir, workdir)

        Utils.mkdir(dir_path, delete_if_exists=True)

        # - Computing optimal header
        logger.info("Mosaicing: computing optimal header ...")

        header_tbl = input_tbl_base + '.hdr'
        header_tbl_fullpath = os.path.join(dir_path, header_tbl)
        montage.mMakeHdr(images_table=input_tbl,
                         template_header=header_tbl_fullpath)

        

        # - Projecting raw frames
        logger.info("Mosaicing: projecting raw frames ...")
        stats_tbl = input_tbl_base + '_stats.tbl'
        stats_tbl_fullpath = os.path.join(dir_path, stats_tbl)

        montage.mProjExec(
            images_table=input_tbl,
            template_header=header_tbl_fullpath,
            proj_dir=dir_path,
            stats_table=stats_tbl_fullpath,
            exact=exact,
            debug=True
        )

        # - List projected frames
        logger.info("Mosaicing: listing projected frames ...")
        projimg_tbl = input_tbl_base + '_proj.tbl'
        projimg_tbl_fullpath = os.path.join(dir_path, projimg_tbl)
        res = montage.mImgtbl(dir_path, projimg_tbl_fullpath)
        if res.count == 0:
            logger.error("Mosaicing: No images were successfully projected!")
            return -1

        # - Compute overlap if background_match is enabled
        if background_match:
            logger.info("Mosaicing: determining overlaps ...")
            diffs_tbl = input_tbl_base + '_diffs.tbl'
            diffs_tbl_fullpath = os.path.join(dir_path, diffs_tbl)
            res = montage.mOverlaps(projimg_tbl_fullpath, diffs_tbl_fullpath)
            if res.count == 0:
                logger.info("Mosaicing: no overlapping frames, backgrounds will not be adjusted")
                background_match = False

        # - Make mosaic
        if background_match:
            logger.error("Mosaicing with background_match not yet implemented!")
            raise NotImplementedError

        else:
            # - Mosaicking frames
            logger.info("Mosaicing: adding frames ...")
            mosaic_file = input_tbl_base + '_mosaic64.fits'
            mosaic_file_fullpath = os.path.join(dir_path, mosaic_file)

        try:
            montage.mAdd(
              projimg_tbl_fullpath,
              header_tbl_fullpath,
              mosaic_file_fullpath,
              img_dir=dir_path,
              type=combine,
              exact=exact
            )
        except:
            logger.error(sys.exc_info())
            return -1


        # - Converting mosaic file to desired format
        logger.info("Mosaicing: converting mosaic file to desired format (type=%s) ..." % (bitpix))


        try:
            montage.mConvert(mosaic_file_fullpath, output, bitpix=bitpix)
        except:
            logger.error(sys.exc_info())
            return -1

        # - Converting mosaic area file to desired format
        mosaic_area_file = input_tbl_base + '_mosaic64_area.fits'
        mosaic_area_file_fullpath = os.path.join(dir_path, mosaic_area_file)
        output_base = Utils.getBaseFileNoExt(output)
        mosaic_area_outfile = output_base + '_area.fits'
        mosaic_area_outfile_fullpath = os.path.join(outdir, mosaic_area_outfile)

        try:
            montage.mConvert(mosaic_area_file_fullpath,mosaic_area_outfile_fullpath, bitpix=bitpix)
        except:
            logger.error(sys.exc_info())
            return -1

        return 0



#!/usr/bin/env python

from __future__ import print_function

##################################################
###          MODULE IMPORT
##################################################
## STANDARD MODULES
import os
import sys
import subprocess
import string
import time
import signal
from threading import Thread
import datetime
import numpy as np
import random
import math
import logging

## COMMAND-LINE ARG MODULES
import getopt
import argparse
import collections

## MODULES
from scutout import __version__, __date__
from scutout import logger
from scutout.config import Config
from scutout.cutout_extractor import CutoutFinder

#### GET SCRIPT ARGS ####
def str2bool(v):
	if v.lower() in ('yes', 'true', 't', 'y', '1'):
		return True
	elif v.lower() in ('no', 'false', 'f', 'n', '0'):
		return False
	else:
		raise argparse.ArgumentTypeError('Boolean value expected.')


###########################
##     ARGS
###########################
def get_args():
	"""This function parses and return arguments passed in"""
	parser = argparse.ArgumentParser(description="Parse args.")
	
	parser.add_argument('-filename','--filename', dest='filename', required=True, type=str, default='', help='List of source position to be searched for cutout (ascii format)') 
	parser.add_argument('-config','--config', dest='config', required=True, type=str, default='', help='Configuration file (INI format)') 
	parser.add_argument('-loglevel','--loglevel', dest='loglevel', required=False, type=str, default='INFO', help='Logging level (default=INFO)') 
	
	# ...
	# ...

	args = parser.parse_args()	

	return args


##############
##   MAIN   ##
##############
def main():
	"""Main function"""

	
	#===========================
	#==   PARSE ARGS
	#===========================
	logger.info("Get script args ...")
	try:
		args= get_args()
	except Exception as ex:
		logger.error("Failed to get and parse options (err=%s)",str(ex))
		return 1

	filename= args.filename
	config_filename= args.config
	log_level_str= args.loglevel.upper()

	log_level = getattr(logging, log_level_str, None)
	if not isinstance(log_level, int):
		logger.error('Invalid log level given: ' + log_level_str)
		return 1
	logger.setLevel(log_level)

	if not filename:
		logger.error("Input file name with coordinate list is empty!")
		return 1
	if not config_filename:
		logger.error("Configuration file name is empty!")
		return 1

	
	#===========================
	#==   PARSE CONFIG
	#===========================
	logger.info("Parse config file ...")

	config= Config()
	status= config.parse(config_filename)	
	if status<0:
		logger.error("Failed to parse and validate config file " + config_filename + "!")
		return 1

	
	#===========================
	#==   RUN CUTOUT SEARCH
	#===========================
	logger.info("Run cutout search ...")
	finder= CutoutFinder(filename,config)

	status= finder.run_search()
	if status<0:
		logger.error("Cutout finder run failed!")
		return -1


	return 0

###################
##   MAIN EXEC   ##
###################
if __name__ == "__main__":
	sys.exit(main())




#!/usr/bin/env python

##################################################
###          MODULE IMPORT
##################################################
## STANDARD MODULES
import os
import sys
import string
import logging
import numpy as np
import errno
import fnmatch
import shutil
import cv2 as cv
import enlighten

## ASTRO MODULES
from astropy.io import fits
from astropy.io import ascii 
from astropy.wcs import WCS
from astropy import units as u
from astropy.convolution import Gaussian2DKernel, convolve
import montage_wrapper as montage
import radio_beam

## GRAPHICS MODULES
#import matplotlib.pyplot as plt

## MODULES
from scutout import logger
from scutout.utils import Utils

logger = logging.getLogger(__name__)


###########################
##     CLASS DEFINITIONS
###########################
class CutoutFinder(object):
	""" Class to extract source cutout from object list file """

	def __init__(self,_filename,_config):
		""" Return a cutout finder object """

		self.filename= _filename
		self.config= _config
		self.table= None
		self.table_size= 0
	
	#==============================
	#     READ INPUT FILE TABLE
	#==============================
	def __read_table(self):
		""" Read file table with object coordinates """	

		# - Read table
		self.table= Utils.read_ascii_table(self.filename,row_start=0,delimiter=' ')
		if not self.table:
			logger.error("Failed to read table!")
			return -1
		
		# - Check table has entries
		self.table_size= len(self.table)
		if self.table_size<=0:
			logger.error("Given input table with coordinates for cutout is empty (hint: check table path & format)!")
			return -1

		# - Print table info
		print(self.table)
		print(self.table.colnames)

		return 0

	#==============================
	#     RUN SEARCH
	#==============================
	def run_search(self):
		""" Run search """

		#**********************
		#     READ TABLE
		#**********************
		if self.__read_table()<0:
			logger.error("Failed to read input table, search failed!")
			return -1

		#**********************
		#     SEARCH CUTOUTS
		#**********************
		has_radius_col= 'RADIUS' in self.table.colnames

		manager = enlighten.get_manager()
		pbar = manager.counter(total=len(self.table), desc='Processing sources', unit='sources')

		for item in self.table:	
			ra= item['RA']
			dec= item['DEC']
			obj_name= item['OBJNAME']
			radius= -1
			if has_radius_col:
				radius= item['RADIUS']
	
			logger.info("Searching cutout for source %s (%f,%f) ..." % (obj_name,ra,dec))
			
			try:
				cs= CutoutHelper(self.config,ra,dec,obj_name,radius)
				status= cs.run()
				if status<0:
					errmsg= 'Failed to extract cutout for source ' + obj_name + ', skip to next...'
					logger.warn(errmsg)
					continue

			except Exception as e:
				#logger.error('Unknown error in source {0}. Trace: {1}'.format(obj_name, e.message))
				logger.error('Unknown error in source {0}. Trace: {1}'.format(obj_name, str(e)))
				continue

			pbar.update()

		return 0


###########################
##     CLASS DEFINITIONS
###########################
class CutoutHelper(object):
	""" Class to extract source cutout from RA/DEC position """

	def __init__(self,_config,_ra,_dec,_obj_name,_radius=-1):
		""" Return a cutout helper object """

		self.config= _config
		self.ra= _ra
		self.dec= _dec
		self.source_radius= _radius
		self.cutout_size= 2*_radius*self.config.cutout_factor
		if _radius<=0 or self.config.use_same_radius:	
			self.source_radius= self.config.source_radius
			self.cutout_size= 2*self.config.source_radius*self.config.cutout_factor
		
		self.bkg_inner_radius= self.source_radius*self.config.bkg_inner_radius_factor # in arcsec
		self.bkg_outer_radius= self.source_radius*self.config.bkg_outer_radius_factor # in arcsec

		self.source_radius/= 3600. # in degrees
		self.cutout_size/= 3600. # in degrees
		

		self.sname= _obj_name
		self.topdir= self.config.workdir + '/' + self.sname
		self.tmpdir= self.topdir + '/tmpfiles'
		self.surveys= self.config.surveys
		self.img_files= {}

	#==============================
	#     INITIALIZE
	#==============================
	def __initialize(self):
		""" Initialize search """

		# - Create source work directories
		logger.info('Creating source work directory ' + self.topdir + ' ...')
		Utils.mkdir(self.tmpdir)
		
		# - Enter tmp directory
		logger.info('Entering source work tmp directory ' + self.tmpdir + ' ...')
		os.chdir(self.tmpdir)

		return 0

	#==============================
	#     EXTRACT RAW CUTOUT
	#==============================
	def __extract_raw_cutout(self,survey):
		""" Find cutout for given survey """
		
		##  Create output directories
		# - Input data
		input_img_dir= self.tmpdir + '/inputs'
		Utils.mkdir(input_img_dir)
		
		# - Raw cutout data
		raw_cutout_dir= self.tmpdir + '/raw_cutouts'	
		Utils.mkdir(raw_cutout_dir) 

		# - Get survey data options		
		survey_opts= self.config.survey_options[survey]
		#print("Survey %s options" % survey)
		#print(survey_opts)
		metadata_tbl= survey_opts['metadata']
		#print("Survey %s metadata=%s" % (survey,metadata_tbl))
		
		# - Search in which survey file the source is located using Montage mCoverageCheck routine
		coverage_tbl= 'coverage_' + survey + '.tbl'
		coverage_tbl_fullpath= self.tmpdir + '/' + coverage_tbl
		logger.info('Making coverage table %s (r=%s arcsec) ...' % (coverage_tbl,str(self.source_radius*3600)))
		#logger.info('Making coverage table ' + coverage_tbl_fullpath + ' (r=' + str(self.source_radius) + ') ...')
		montage.mCoverageCheck(
			in_table=metadata_tbl,
			out_table=coverage_tbl_fullpath,
			mode='circle',
			ra=self.ra, dec=self.dec,
			radius=self.source_radius
		)

		# - Read coverage table to check if an image was found
		try:
			table= ascii.read(coverage_tbl_fullpath)
		except Exception as ex:
			logger.error('Failed to read coverage table!')
			return -1

		nimgs= len(table)
		if nimgs<=0:
			logger.info("No survey images found covering given source coordinates, go to next survey data...")
			return 0

		imgfile_fullpath= table[0]['fname']		

		if nimgs>1:
			logger.info("More than 1 image found covering source coordinates, using mode %s ..." % self.config.multi_input_img_mode)
			if self.config.multi_input_img_mode=='first':
				imgfile_fullpath= table[0]['fname']

			elif self.config.multi_input_img_mode=='best':
				try:
					res= montage.mBestImage(images_table=coverage_tbl_fullpath,ra=self.ra,dec=self.dec)
					imgfile_fullpath= res.file
				except:
					logger.warn("Caught exception from Montage mBestImage, fallback to first image ...")
					imgfile_fullpath= table[0]['fname']				

			elif self.config.multi_input_img_mode=='mosaic':
				mosaic_file= 'mosaic_' + survey + '.fits'
				mosaic_file_fullpath= os.path.join(input_img_dir,mosaic_file)
				status= Utils.makeMosaic(coverage_tbl_fullpath,output=mosaic_file_fullpath)	
				if status==0:
					imgfile_fullpath= mosaic_file_fullpath
				else:
					logger.warn("Failed to compute mosaic from images listed in file %s, taking best one out of them ..." % (coverage_tbl))
					try:
						res= montage.mBestImage(images_table=coverage_tbl_fullpath,ra=self.ra,dec=self.dec)
						imgfile_fullpath= res.file
					except:
						logger.warn("Caught exception from Montage mBestImage, fallback to first image ...")
						imgfile_fullpath= table[0]['fname']			
			else:
				logger.warn("Invalid/unknown multi input image option (%s), taking the first one..." % (self.config.multi_input_img_mode))
				imgfile_fullpath= table[0]['fname']
		
			
		#imgfile= table[0]['fname']
		#imgfile_fullpath= imgfile
		#imgfile_base= os.path.basename(imgfile_fullpath)
		imgfile_local= self.sname + '_' + survey + '.fits'
		imgfile_local_fullpath= self.tmpdir + '/' + imgfile_local
		
		# - Copy input file to tmp dir
		#shutil.copy(imgfile_fullpath,imgfile_local_fullpath)

		# - Extract the cutout using Montage
		logger.info("Extracting raw cutout around (%s,%s) with size %s (arcsec) ..." % (str(self.ra),str(self.dec),str(self.cutout_size*3600)))
		cutout_file= self.sname + '_' + survey + '_cut.fits'
		cutout_file_fullpath= self.tmpdir + '/' + cutout_file
		raw_cutout_file= ''
		raw_cutout_file_fullpath= ''
		montage.mSubimage(
			#in_image=imgfile_local_fullpath, 
			in_image=imgfile_fullpath, 
			out_image=cutout_file_fullpath, 
			ra=self.ra, dec=self.dec, xsize=self.cutout_size
		)
		self.img_files[survey]= cutout_file_fullpath
		
		# - Fix image axis or scale in case BZERO!=0 or BSCALE!=0
		logger.info("Fixing possible degenerate axis and non-null BSCALE factor in image %s ..." % (cutout_file))
		status= Utils.fixImgAxisAndUnits(cutout_file_fullpath,cutout_file_fullpath)
		if status<0:
			logger.error("Failed to adjust axis/scale of image " + cutout_file_fullpath + "!")
			return -1

		# - Convert to Jy/pixel units?
		#band= -1
		#pos= survey.find('_b')
		#if pos>0:
		#	band= int(survey[pos+2:])

		if self.config.convert_to_jypix_units:
			cutout_file_scaled= self.sname + '_' + survey + '_cut_jy.fits'
			cutout_file_scaled_fullpath= self.tmpdir + '/' + cutout_file_scaled
			raw_cutout_file= cutout_file
			raw_cutout_file_fullpath= cutout_file_fullpath

			logger.info('Converting image %s in Jy/pixel units ...' % (raw_cutout_file))
			Utils.convertImgToJyPixel(cutout_file_fullpath,cutout_file_scaled_fullpath,survey,table[0]['bunit'])
			self.img_files[survey]= cutout_file_scaled_fullpath
			
		
		## Organize files in directories or remove some of them
		# - Move coverage table to input subdir
		shutil.move(coverage_tbl_fullpath,os.path.join(input_img_dir,coverage_tbl))
		
		# -  Move raw cutout file to cutout subdir
		if raw_cutout_file:
			logger.debug("Moving file " + raw_cutout_file + ' to ' + raw_cutout_dir + ' ...')
			shutil.move(raw_cutout_file_fullpath,os.path.join(raw_cutout_dir,raw_cutout_file))

		# - Move final products in subdir
		shutil.move(self.img_files[survey],os.path.join(raw_cutout_dir,os.path.basename(self.img_files[survey])))
		self.img_files[survey]= os.path.join(raw_cutout_dir,os.path.basename(self.img_files[survey]))		

		#print("Cutout images to be processed after raw cutout step")
		#print(self.img_files)

		return 0

	#==============================
	#     SUBTRACT BKG
	#==============================
	def __subtract_bkg(self):
		""" Subtract bkg from raw cutout """

		# - Create output directory
		bkgsub_cutout_dir= self.tmpdir + '/bkgsub_cutouts'	
		Utils.mkdir(bkgsub_cutout_dir)

		#	- Loop over cutouts and compute bkg
		for survey, filename in self.img_files.items(): 
			filename_base= Utils.getBaseFileNoExt(filename)
			bkgsub_cutout= filename_base + '_bkgsub.fits'
			bkgsub_cutout_fullpath= self.tmpdir + '/' + bkgsub_cutout
			self.img_files[survey]= bkgsub_cutout_fullpath
	
			# - Compute bkg in annulus around source
			logger.info("Computing bkg for image %s (R1=%s, R2=%s) ..." % (filename_base,str(self.bkg_inner_radius),str(self.bkg_outer_radius)))			
			bkg= Utils.estimateBkgFromAnnulus(
				filename=filename,
				ra=self.ra,dec=self.dec,
				R1=self.bkg_inner_radius,R2=self.bkg_outer_radius,
				method=self.config.bkg_estimator,
				max_nan_thr=self.config.bkg_max_nan_thr
			)

			# - Read fits image
			data, header= Utils.read_fits(filename)
	
			# - Subtract bkg?
			good_bkg= (bkg>0 and np.isfinite(bkg))
			if good_bkg:
				logger.info("Subtracting bkg=%s from image %s ..." % (str(bkg),filename_base))			
				data_nobkg= data - bkg
			else:
				logger.warn("Computed bkg is not valid (negative/nan/inf), won't subtract bkg from image %s ..." % (filename_base))
				data_nobkg= data			

			# - Write fits
			logger.debug("Writing bkg-subtracted image to file %s ..." % (bkgsub_cutout))
			Utils.write_fits(data_nobkg,bkgsub_cutout_fullpath,header)

		# - Move final products in subdir
		for survey, path in self.img_files.items(): 
			shutil.move(self.img_files[survey],os.path.join(bkgsub_cutout_dir,os.path.basename(self.img_files[survey])))
			self.img_files[survey]= os.path.join(bkgsub_cutout_dir,os.path.basename(self.img_files[survey]))


		return 0

	#==============================
	#     REGRID CUTOUTS
	#==============================
	def __regrid_cutouts(self):
		""" Regrid cutouts to the same projection and pixel """

		##  Create output directories
		# - Raw cutout dir (should already exist if previous step was called)
		raw_cutout_dir= self.tmpdir + '/raw_cutouts'	
		Utils.mkdir(raw_cutout_dir)

		# - Regrid cutout dir
		reproj_cutout_dir= self.tmpdir + '/reproj_cutouts'	
		Utils.mkdir(reproj_cutout_dir)

		# - Get list of raw cutout images to be re-projected
		raw_cutouts= []
		reproj_cutouts= []
		dx_orig= []
		dy_orig= []

		for survey, path in self.img_files.items(): 
			reproj_cutout= Utils.getBaseFileNoExt(path) + '_reproj.fits'
			reproj_cutout_fullpath= self.tmpdir + '/' + reproj_cutout
			raw_cutouts.append(path)
			reproj_cutouts.append(reproj_cutout_fullpath)
			self.img_files[survey]= reproj_cutout_fullpath

		#print(raw_cutouts)
		#print(reproj_cutouts)

		# - Reproject cutouts using py Montage reproject high-level API
		try:
			montage.reproject(
				in_images=raw_cutouts, 
				out_images=reproj_cutouts, 
				header=None, 
				bitpix=None, 
				north_aligned=True,
				common=True	
			)
		except Exception as e:
			logger.error("Failed to reproject cutouts (err=%s)!" % str(e))
			return -1
			

		# - Scale image data to conserve flux, e.g. multiply data by (pix1/pix1_ori)*(pix2/pix2_ori)
		# - Copy back beam information (montage does not include in the re-projected image file)
		# - Overwrite previous image
		
		for index in range(len(raw_cutouts)):
			data, header= Utils.read_fits(raw_cutouts[index])
			dx= abs(header['CDELT1'])
			dy= abs(header['CDELT2'])
		
			data_reproj, header_reproj= Utils.read_fits(reproj_cutouts[index])
			dx_reproj= abs(header_reproj['CDELT1'])
			dy_reproj= abs(header_reproj['CDELT2'])

			flux_scale= (dx_reproj/dx)*(dy_reproj/dy)
			data_reproj_scaled= flux_scale*data_reproj
	
			logger.info("Scaling re-projected image %s to conserve flux (conv factor=%s) ..." % (raw_cutouts[index],str(flux_scale)))

			if Utils.hasBeamInfo(header) and not Utils.hasBeamInfo(header_reproj):
				header_reproj['BMAJ']= header['BMAJ']
				header_reproj['BMIN']= header['BMIN']
				header_reproj['BPA']= header['BPA'] if 'BPA' in header else 0.

			Utils.write_fits(data_reproj_scaled,reproj_cutouts[index],header_reproj)
			

		# - Move final products to subdir
		for survey, path in self.img_files.items(): 
			shutil.move(self.img_files[survey],os.path.join(reproj_cutout_dir,os.path.basename(self.img_files[survey])))
			self.img_files[survey]= os.path.join(reproj_cutout_dir,os.path.basename(self.img_files[survey]))
		
		#print("Cutout images to be processed after cutout reproj step")
		#print(self.img_files)

		return 0


	#=========================================
	#     CONVOLVE CUTOUTS TO SAME RESOLUTION
	#=========================================
	def __convolve_to_same_resolution(self):
		""" Convolve cutouts to same resolution """
		
		## Create output directory
		conv_cutout_dir= self.tmpdir + '/conv_cutouts'	
		Utils.mkdir(conv_cutout_dir)
		
		# - Get list of raw cutout images to be convolved
		raw_cutouts= []
		conv_cutouts= []
		beam_list= []
		pixsize_x= []
		pixsize_y= []
		data_list= []
		header_list= []

		for survey, path in self.img_files.items(): 
			raw_cutouts.append(path)
			conv_cutout= Utils.getBaseFileNoExt(path) + '_conv.fits'
			conv_cutout_fullpath= self.tmpdir + '/' + conv_cutout
			conv_cutouts.append(conv_cutout_fullpath)
			self.img_files[survey]= conv_cutout_fullpath
			
			# - Get image beam
			data, header= Utils.read_fits(path)
			wcs = WCS(header)
			data_list.append(data)
			header_list.append(header)

			hasBeamInfo= Utils.hasBeamInfo(header)
			xc= header['CRPIX1']
			yc= header['CRPIX2']	
			ra, dec = wcs.all_pix2world(xc,yc,0,ra_dec_order=True)
			dx= abs(header['CDELT1']) # in deg
			dy= abs(header['CDELT2']) # in deg
			pixsize_x.append(dx)
			pixsize_y.append(dy)

			if hasBeamInfo:
				bmaj= header['BMAJ'] # in deg
				bmin= header['BMIN'] # in deg
				pa= header['BPA'] if 'BPA' in header else 0 # in deg
				beam= radio_beam.Beam(bmaj*u.deg,bmin*u.deg,pa*u.deg)
			else:
				logger.warn("No BMAJ/BMIN keyword present in file " + path + ", trying to retrieve from survey name...")
				beamArea= Utils.getSurveyBeamArea(survey,ra,dec)
				bmaj= np.sqrt(beamArea*4.*np.log(2)/np.pi)
				if beamArea>0:
					beam= radio_beam.Beam(bmaj*u.deg,bmaj*u.deg)
				else:
					logger.error("No BMAJ keyword present in file " + path + ", cannot compute conversion factor!")
					return -1

			beam_list.append(beam)
		
		# - Compute common beam
		beams= radio_beam.Beams(beams=beam_list)
		common_beam= radio_beam.commonbeam.common_manybeams_mve(beams)
		common_beam_bmaj= common_beam.major.to(u.arcsec).value
		common_beam_bmin= common_beam.minor.to(u.arcsec).value
		common_beam_pa= common_beam.pa.to(u.deg).value
		

		# - Loop over image, find conv beam to be used to reach common beam and convolve image with this
		logger.info("Convolving images to common beam size (bmaj,bmin,pa)=(%s,%s,%s) ..." % (str(common_beam_bmaj),str(common_beam_bmin),str(common_beam_pa)))
		
		for index in range(len(raw_cutouts)):

			# - Find convolving beam 
			logger.debug("Finding convolving beam ...")
			try:
				bmaj, bmin, pa= radio_beam.utils.deconvolve(common_beam,beam_list[index])
			except Exception as e:	
				logger.error("Failed to deconvolve beam no. %d (err=%s)" % (index,str(e)))
				continue

			logger.debug("type(bmaj)")
			logger.debug(type(bmaj))
			logger.debug("type(bmin)")
			logger.debug(type(bmin))
			logger.debug("type(pa)")
			logger.debug(type(pa))

			try:
				#bmaj_deg= bmaj.to(u.deg).value
				#bmaj_arcsec= bmaj.to(u.arcsec).value
				bmaj_deg= bmaj.to_value(u.deg)
				bmaj_arcsec= bmaj.to_value(u.arcsec)
			except Exception as e:
				logger.warn("Failed to convert bmaj to no unit values (possibly not astropy Units type) (err=%s), assuming it is a scalar ..." % str(e))
				bmaj_deg= bmaj
				bmaj_arcsec= bmaj*3600
				
			try:
				#bmin_deg= bmin.to(u.deg).value
				#bmin_arcsec= bmin.to(u.arcsec).value
				bmin_deg= bmin.to_value(u.deg)
				bmin_arcsec= bmin.to_value(u.arcsec)
			except Exception as e:
				logger.warn("Failed to convert bmin to no unit values (possibly not astropy Units type) (err=%s), assuming it is a scalar ..." % str(e))
				bmin_deg= bmin
				bmin_arcsec= bmin*3600

			try:
				#pa_deg= pa.to(u.deg).value
				pa_deg= pa.to_value(u.deg)
			except Exception as e:
				logger.debug("Failed to convert pa to no unit values (possibly not astropy Units type) (err=%s), assuming it is a scalar ..." % str(e))
				pa_deg= pa
				

			logger.debug("Creating radio beam object ...")
			logger.debug("type(bmaj_deg)")
			logger.debug(type(bmaj_deg))

			try:
				conv_beam= radio_beam.Beam(bmaj_deg*u.deg,bmin_deg*u.deg,pa_deg*u.deg)
			except Exception as e:	
				logger.error("Failed to create conv beam from (bmaj,bmin,pa)=(%f,%f,%f) (err=%s)" % (bmaj_deg,bmin_deg,pa_deg,str(e)))
				continue			

			ny= data_list[index].shape[0]
			nx= data_list[index].shape[1]
			
			
			# - Create convolution kernel
			logger.info("Convolving image %s (size=%d,%d) with beam (bmaj,bmin,pa)=(%f,%f,%f) ..." % (raw_cutouts[index],nx,ny,bmaj_arcsec,bmin_arcsec,pa_deg))
			dx= pixsize_x[index]
			dy= pixsize_y[index]
			pixsize= max(dx,dy)
			conv_kernel= conv_beam.as_kernel(pixsize*u.deg)
			conv_kernel.normalize()
			kernel=conv_kernel.array

			# - Convolve image and write fits
			data_list[index][np.isnan(data_list[index])]=0.0
			data_conv=cv.filter2D(np.float64(data_list[index]),-1,kernel,borderType=cv.BORDER_CONSTANT)
			logger.info("Kernel size: %d x %d" % (kernel.shape[0],kernel.shape[1]))
			Utils.write_fits(data_conv,conv_cutouts[index],header_list[index])

		
		# - Move final products in subdir
		for survey, path in self.img_files.items(): 
			shutil.move(self.img_files[survey],os.path.join(conv_cutout_dir,os.path.basename(self.img_files[survey])))
			self.img_files[survey]= os.path.join(conv_cutout_dir,os.path.basename(self.img_files[survey]))

		#print("Cutout images to be processed after cutout convolve step")
		#print(self.img_files)


		return 0


	#==============================
	#     CROP CUTOUTS
	#==============================
	def __crop(self):
		""" Crop cutouts to the desired number of pixels """

		## Organize files in directories 
		crop_cutout_dir= self.tmpdir + '/cropped_cutouts'	
		Utils.mkdir(crop_cutout_dir)
		
		# - Computing source size. Crop method will internally check is cutout size is cutting part of the source
		source_size= self.source_radius # in deg

		for survey, filename in self.img_files.items(): 
			# - Set output filename
			cropped_cutout= Utils.getBaseFileNoExt(filename) + '_cropped.fits'
			cropped_cutout_fullpath= self.tmpdir + '/' + cropped_cutout
			#cropped_cutouts.append(cropped_cutout_fullpath)
			self.img_files[survey]= cropped_cutout_fullpath

			# - Crop image
			logger.info("Cropping cutout file " + filename + " to desired size ...")
			status= Utils.cropImage(
				filename=filename,
				ra=self.ra,dec=self.dec,
				crop_mode = self.config.crop_mode,
				crop_size = self.config.crop_size,
				outfile=cropped_cutout_fullpath,
				source_size=source_size,
				nanfill=True,
				nanfill_mode='imgmin',
				nanfill_val=0
			)

			# - Move conv cutout in subdir
			#filename_base= Utils.getBaseFile(filename)	
			#logger.info("Moving file " + filename_base + ' to ' + conv_cutout_dir + ' ...')
			#shutil.move(filename,os.path.join(conv_cutout_dir,filename_base))
	
			# - Exit if crop failed
			if status<0:
				logger.error("Failed to crop cutout file " + filename + " (hints: check is source is larger than crop size)")
				return -1

			# - Copy final product in subdir
			logger.debug("Moving file " + os.path.basename(self.img_files[survey]) + ' to ' + crop_cutout_dir + ' ...')
			shutil.move(self.img_files[survey],os.path.join(crop_cutout_dir,os.path.basename(self.img_files[survey])))
			self.img_files[survey]= os.path.join(crop_cutout_dir,os.path.basename(self.img_files[survey]))



		return 0
		
	#==============================
	#     RUN CUTOUT EXTRACTION
	#==============================
	def run(self):
		""" Run search for single source """
		
		#**********************
		#        INIT
		#**********************
		if self.__initialize()<0:
			logger.error("Failed to initialize source cutout search!")
			return -1
		
		#*************************************
		#   EXTRACT CUTOUT FROM SURVEY DATA
		#*************************************
		for survey in self.surveys:
			logger.info('Searching cutout for source ' + self.sname + ' for survey ' + survey + ' ...')
			status= self.__extract_raw_cutout(survey)
			if status<0:
				logger.error('Raw cutout extraction for source ' + self.sname + ' for survey ' + survey + ' failed!')
				return -1

		#*************************************
		#   SUBTRACT BACKGROUND
		#*************************************
		if self.config.subtract_bkg:
			logger.info('Subtracting bkg from raw cutouts ...')
			status= self.__subtract_bkg()
			if status<0:
				logger.error('Failed to subtract bkg from raw cutouts for source ' + self.sname + '!')
				return -1

		#*************************************
		#        REGRID CUTOUTS
		#*************************************
		if self.config.regrid:
			logger.info('Regridding raw cutouts ...')
			status= self.__regrid_cutouts()
			if status<0:
				logger.error('Failed to regrid raw cutouts for source ' + self.sname + '!')
				return -1

		#*************************************
		#       CONVOLVE CUTOUTS
		#*************************************
		if self.config.convolve:
			logger.info('Convolving raw cutouts to the same resolution ...')
			status= self.__convolve_to_same_resolution()
			if status<0:
				logger.error('Failed to convolve raw cutouts for source ' + self.sname + '!')
				return -1

		#*************************************
		#       CROP CUTOUTS
		#*************************************
		if self.config.crop_mode!='none':
			logger.info('Cropping cutouts in ' + self.config.crop_mode +  ' mode')
			status= self.__crop()
			if status<0:
				logger.error('Failed to crop cutouts for source ' + self.sname + '!')
				return -1
	
		#*************************************
		#     COPY FINAL FILES IN MAIN DIR
		#*************************************
		# - Copy final files in main directory
		logger.debug("Copying final image cutouts in main directory ...")
		for survey, filename in self.img_files.items():
			filename_base= Utils.getBaseFile(filename)
			filename_final= self.sname + '_' + survey + '.fits'
			filename_final_fullpath= self.topdir + '/' + filename_final
			shutil.copy(filename,filename_final_fullpath)

		# - Remove tmp file directory?
		if not self.config.keep_tmpfiles:
			logger.debug("Removing tmp file dir " + self.tmpdir + " ...")
			shutil.rmtree(self.tmpdir, ignore_errors=True)

		return 0




#! /usr/bin/env python

__title__ = 'scutout'
__version__ = '1.0.0'
__author__ = 'Simone Riggi'
__license__ = 'GPL3'
__date__ = '2019-05-29'
__copyright__ = 'Copyright 2019 by Simone Riggi - INAF'


import logging
import logging.config


# Create the Logger
#logging.basicConfig(format="%(module)s:%(levelname)s %(message)s")
logging.basicConfig(format="%(asctime)-15s %(levelname)s - %(message)s",datefmt='%Y-%m-%d %H:%M:%S')
logger= logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.info("This is scutout {0}-({1})".format(__version__, __date__))





#! /usr/bin/env python
"""
Setup for scutout
"""
import os
import sys
from setuptools import setup


def read(fname):
	"""Read a file"""
	return open(os.path.join(os.path.dirname(__file__), fname)).read()


def get_version():
	""" Get the package version number """
	import scutout
	return scutout.__version__



PY_MAJOR_VERSION=sys.version_info.major
PY_MINOR_VERSION=sys.version_info.minor
print("PY VERSION: maj=%s, min=%s" % (PY_MAJOR_VERSION,PY_MINOR_VERSION))

reqs= []
reqs.append('numpy>=1.10')
reqs.append('astropy>=2.0, <3')
reqs.append('montage-wrapper')
reqs.append('radio-beam')
reqs.append('regions')

if PY_MAJOR_VERSION<=2:
	print("PYTHON 2 detected")
	reqs.append('future')
	reqs.append('scipy<=1.2.1')
	reqs.append('pyparsing>=2.0.1')
	#reqs.append('matplotlib<=2.2.4')
else:
	print("PYTHON 3 detected")
	reqs.append('scipy')
	reqs.append('pyparsing')
	#reqs.append('matplotlib')




data_dir = 'data'

setup(
	name="scutout",
	version=get_version(),
	author="Simone Riggi",
	author_email="simone.riggi@gmail.com",
	description="Tool to extract source cutouts from a collection of astronomical FITS images",
	license = "GPL3",
	url="https://github.com/SKA-INAF/scutout",
  download_url="https://github.com/user/reponame/archive/v1.0.0.tar.gz",
	keywords = ['source', 'cutout', 'survey'],
	long_description=read('README.md'),
	long_description_content_type='text/markdown',
	packages=['scutout'],
	install_requires=reqs,
	scripts=['scripts/run_scutout.py'],
	classifiers=[
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Science/Research',
    'Topic :: Scientific/Engineering :: Astronomy',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Programming Language :: Python :: 3'
  ]
)



#!/usr/bin/env python

##################################################
###          MODULE IMPORT
##################################################
## STANDARD MODULES
import os
import sys
import string
import logging
import io

## CONFIG PARSER MODULES
try:
	from ConfigParser import SafeConfigParser
	import ConfigParser
except:
	# in Python3 it was renamed from ConfigParser to configparser
	from configparser import SafeConfigParser
	import configparser

##############################
##     GLOBAL VARS
##############################
logger = logging.getLogger(__name__)

###########################
##     CLASS DEFINITIONS
###########################
class CaseConfigParser(SafeConfigParser):
	def optionxform(self, optionstr):
		return optionstr

class Config(object):
     
	""" Configuration options """

	def __init__(self):
		""" Return a Config object """

		# - Config options
		self.__set_defaults()

		# - Config file parser
		try: # py2
			self.parser= ConfigParser.ConfigParser()
		except: # py3
			self.parser= configparser.ConfigParser()

		self.config= None

	def optionxform(self, optionstr):
		return optionstr

	#==============================
	#     SET DEFAULT OPTIONS
	#==============================
	def __set_defaults(self):
		""" Set default options """
				
		# - Run options		
		self.workdir= os.getcwd()
		self.keep_tmpfiles= True
		
		# - Cutout search
		self.surveys= []
		self.use_same_radius= False
		self.source_radius= 300 # arcsec
		self.cutout_factor= 5 
		self.multi_input_img_mode= 'best'
		self.convert_to_jypix_units= True
		self.subtract_bkg= False
		self.regrid= True
		self.convolve= True
		self.crop_mode= 'none'
		self.crop_size= 200 # in pixels

		# - Background calculation
		self.bkg_estimator= 'sigmaclip'
		self.bkg_inner_radius_factor= 1.1
		self.bkg_outer_radius_factor= 1.2
		self.bkg_max_nan_thr= 0.1


		# - FIRST survey options
		first_options= {
			"metadata": ""
		}

		# - NVSS survey options
		nvss_options= {
			"metadata": ""
		}

		# - MGPS survey options
		mgps_options= {
			"metadata": ""
		}

		# - VGPS survey options
		vgps_options= {
			"metadata": ""
		}

		# - SGPS survey options
		sgps_options= {
			"metadata": ""
		}

		# - VLASS survey options
		vlass_options= {
			"metadata": ""
		}

		# - SCORPIO ATCA survey options
		scorpio_atca_2_1_options= {
			"metadata": ""
		}

		# - SCORPIO ASKAP survey options
		scorpio_askap15_b1_options= {
			"metadata": ""
		}
		scorpio_askap36_b123_options= {
			"metadata": ""
		}
		scorpio_askap36_b123_ch1_options= {
			"metadata": ""
		}
		scorpio_askap36_b123_ch2_options= {
			"metadata": ""
		}
		scorpio_askap36_b123_ch3_options= {
			"metadata": ""
		}
		scorpio_askap36_b123_ch4_options= {
			"metadata": ""
		}
		scorpio_askap36_b123_ch5_options= {
			"metadata": ""
		}
		
		# - THOR survey options
		thor_options= {
			"metadata": ""
		}

		# - MEERKAT-GPS survey options
		meerkat_gps_options= {
			"metadata": ""
		}
		meerkat_gps_ch1_options= {
			"metadata": ""
		}
		meerkat_gps_ch2_options= {
			"metadata": ""
		}
		meerkat_gps_ch3_options= {
			"metadata": ""
		}
		meerkat_gps_ch4_options= {
			"metadata": ""
		}
		meerkat_gps_ch5_options= {
			"metadata": ""
		}
		meerkat_gps_ch6_options= {
			"metadata": ""
		}
		meerkat_gps_ch7_options= {
			"metadata": ""
		}
		meerkat_gps_ch8_options= {
			"metadata": ""
		}
		meerkat_gps_ch9_options= {
			"metadata": ""
		}
		meerkat_gps_ch10_options= {
			"metadata": ""
		}
		meerkat_gps_ch11_options= {
			"metadata": ""
		}
		meerkat_gps_ch12_options= {
			"metadata": ""
		}
		meerkat_gps_ch13_options= {
			"metadata": ""
		}
		meerkat_gps_ch14_options= {
			"metadata": ""
		}

		# - ASKAP RACS survey options
		askap_racs_options= {
			"metadata": ""
		}

		# - MAGPIS 21cm survey options
		magpis_21cm_options= {
			"metadata": ""
		}

		# - CORNISH 5 GHz survey options
		cornish_options= {
			"metadata": ""
		}

		# - Spitzer IRAC survey options
		spitzer_irac_3_6_options= {
			"metadata": ""
		}
		spitzer_irac_4_5_options= {
			"metadata": ""
		}
		spitzer_irac_5_8_options= {
			"metadata": ""
		}
		spitzer_irac_8_options= {
			"metadata": ""
		}

		# - Spitzer MIPS 24
		spitzer_mips_24_options= {
			"metadata": ""
		}

		# - Herschel HIGAL 
		herschel_higal_70_options= {
			"metadata": ""
		}
		herschel_higal_160_options= {
			"metadata": ""
		}
		herschel_higal_250_options= {
			"metadata": ""
		}
		herschel_higal_350_options= {
			"metadata": ""
		}
		herschel_higal_500_options= {
			"metadata": ""
		}
		
		# - WISE 
		wise_3_4_options= {
			"metadata": ""
		}
		wise_4_6_options= {
			"metadata": ""
		}
		wise_12_options= {
			"metadata": ""
		}
		wise_22_options= {
			"metadata": ""
		}

		# - APEX ATLAS GAL
		apex_atlasgal_options= {
			"metadata": ""
		}
		apex_atlasgal_planck_options= {
			"metadata": ""
		}

		# - MSX
		msx_8_3_options= {
			"metadata": ""
		}
		msx_12_1_options= {
			"metadata": ""
		}
		msx_14_7_options= {
			"metadata": ""
		}	
		msx_21_3_options= {
			"metadata": ""
		}

		# - CUSTOM survey (placeholder for new survey data added by users)
		custom_survey_options= {
			"metadata": ""
		}

	
		# - SURVEY options
		self.survey_options= {
			"first" : first_options,
			"nvss" : nvss_options,
			"mgps" : mgps_options,
			"vgps" : vgps_options,
			"sgps" : sgps_options,
			"vlass" : vlass_options,
			"scorpio_atca_2_1" : scorpio_atca_2_1_options,
			"scorpio_askap15_b1" : scorpio_askap15_b1_options,
			"scorpio_askap36_b123" : scorpio_askap36_b123_options,
			"scorpio_askap36_b123_ch1" : scorpio_askap36_b123_ch1_options,
			"scorpio_askap36_b123_ch2" : scorpio_askap36_b123_ch2_options,
			"scorpio_askap36_b123_ch3" : scorpio_askap36_b123_ch3_options,
			"scorpio_askap36_b123_ch4" : scorpio_askap36_b123_ch4_options,
			"scorpio_askap36_b123_ch5" : scorpio_askap36_b123_ch5_options,
			"thor" : thor_options,
			"meerkat_gps" : meerkat_gps_options,
			"meerkat_gps_ch1" : meerkat_gps_ch1_options,
			"meerkat_gps_ch2" : meerkat_gps_ch2_options,
			"meerkat_gps_ch3" : meerkat_gps_ch3_options,
			"meerkat_gps_ch4" : meerkat_gps_ch4_options,
			"meerkat_gps_ch5" : meerkat_gps_ch5_options,
			"meerkat_gps_ch6" : meerkat_gps_ch6_options,
			"meerkat_gps_ch7" : meerkat_gps_ch7_options,
			"meerkat_gps_ch8" : meerkat_gps_ch8_options,
			"meerkat_gps_ch9" : meerkat_gps_ch9_options,
			"meerkat_gps_ch10" : meerkat_gps_ch10_options,
			"meerkat_gps_ch11" : meerkat_gps_ch11_options,
			"meerkat_gps_ch12" : meerkat_gps_ch12_options,
			"meerkat_gps_ch13" : meerkat_gps_ch13_options,
			"meerkat_gps_ch14" : meerkat_gps_ch14_options,
			"askap_racs" : askap_racs_options,
			"magpis_21cm" : magpis_21cm_options,
			"cornish" : cornish_options,
			"irac_3_6" : spitzer_irac_3_6_options,
			"irac_4_5" : spitzer_irac_4_5_options,
			"irac_5_8" : spitzer_irac_5_8_options,
			"irac_8" : spitzer_irac_8_options,
			"mips_24" : spitzer_mips_24_options,
			"higal_70" : herschel_higal_70_options,
			"higal_160" : herschel_higal_160_options,
			"higal_250" : herschel_higal_250_options,
			"higal_350" : herschel_higal_350_options,
			"higal_500" : herschel_higal_500_options,
			"wise_3_4" : wise_3_4_options,
			"wise_4_6" : wise_4_6_options,
			"wise_12" : wise_12_options,
			"wise_22" : wise_22_options,
			"atlasgal" : apex_atlasgal_options,
			"atlasgal_planck" : apex_atlasgal_planck_options,
			"msx_8_3" : msx_8_3_options,
			"msx_12_1" : msx_12_1_options,
			"msx_14_7" : msx_14_7_options,
			"msx_21_3" : msx_21_3_options,
			"custom_survey": custom_survey_options,
		}

	#==============================
	#     READ CONFIG FILE
	#==============================
	def __read(self, filename):
		""" Read input INI config file"""

		try:
			self.parser.read(filename)
		except Exception as e:
			logger.error("Exception occurred when reading config file %s (err=%s)!" % (filename, str(e)))
			return -1
		
		#print(self.parser.sections())
		#self.__print_options()

		return 0

	#==============================
	#     ADD CUSTOM SURVEY
	#==============================
	def __add_custom_survey(self, metadata_path):
		""" Add custom survey to config parser (to be called after read config) """
	
		# - Check if metadata field is empty
		if metadata_path=="":
			logger.error("Empty metadata path string given!")
			return -1	

		# - Add section if not already present
		section_name= "CUSTOM_SURVEY_DATA"
		if not self.parser.has_section(section_name):
			self.parser.add_section(section_name)

		# - Add metadata field to section
		self.parser.set(section_name, 'metadata', metadata)

		return 0

	#==============================
	#     PARSE CONFIG FILE
	#==============================
	def __set_options(self):
		""" Parse input INI config file after reading it """

		# - Parse RUN section options
		if self.parser.has_option('RUN', 'workdir'):
			option_value= self.parser.get('RUN', 'workdir')	
			if option_value:
				self.workdir= option_value
		
		if self.parser.has_option('RUN', 'keep_tmpfiles'):
			self.keep_tmpfiles= self.parser.getboolean('RUN', 'keep_tmpfiles')		
		#if self.parser.has_option('RUN', 'keep_inputs'):
		#	self.keep_inputs= self.parser.getboolean('RUN', 'keep_inputs')
		#if self.parser.has_option('RUN', 'keep_tmpcutouts'):
		#	self.keep_tmpcutouts= self.parser.getboolean('RUN', 'keep_tmpcutouts')
		
		# - Parse cutout option sections
		if self.parser.has_option('CUTOUT_SEARCH', 'surveys'):
			option_value= self.parser.get('CUTOUT_SEARCH', 'surveys')	
			if option_value:
				self.surveys= option_value.split(",")
		
		if self.parser.has_option('CUTOUT_SEARCH', 'use_same_radius'):
			self.use_same_radius= self.parser.getboolean('CUTOUT_SEARCH', 'use_same_radius') 		
	
		if self.parser.has_option('CUTOUT_SEARCH', 'source_radius'):
			option_value= self.parser.get('CUTOUT_SEARCH', 'source_radius')
			if option_value:
				self.source_radius= float(option_value)

		if self.parser.has_option('CUTOUT_SEARCH', 'cutout_factor'):
			option_value= self.parser.get('CUTOUT_SEARCH', 'cutout_factor')
			if option_value:
				self.cutout_factor= float(option_value)	
		
		if self.parser.has_option('CUTOUT_SEARCH', 'multi_input_img_mode'):
			option_value= self.parser.get('CUTOUT_SEARCH', 'multi_input_img_mode')	
			if option_value:
				self.multi_input_img_mode= option_value
		
		if self.parser.has_option('CUTOUT_SEARCH', 'convert_to_jy_pixel'):
			self.convert_to_jypix_units= self.parser.getboolean('CUTOUT_SEARCH', 'convert_to_jy_pixel') 		
		
		if self.parser.has_option('CUTOUT_SEARCH', 'subtract_bkg'):
			self.subtract_bkg= self.parser.getboolean('CUTOUT_SEARCH', 'subtract_bkg') 	

		if self.parser.has_option('CUTOUT_SEARCH', 'regrid'):
			self.regrid= self.parser.getboolean('CUTOUT_SEARCH', 'regrid') 		

		if self.parser.has_option('CUTOUT_SEARCH', 'convolve'):
			self.convolve= self.parser.getboolean('CUTOUT_SEARCH', 'convolve') 
		
		if self.parser.has_option('CUTOUT_SEARCH', 'crop_mode'):
			option_value= self.parser.get('CUTOUT_SEARCH', 'crop_mode')
			if option_value:
				self.crop_mode = option_value

		if self.parser.has_option('CUTOUT_SEARCH', 'crop_size'):
			option_value= self.parser.get('CUTOUT_SEARCH', 'crop_size')
			if option_value:
				self.crop_size= float(option_value)

		# - Parse background options
		if self.parser.has_option('BKG_SUBTRACTION', 'bkg_estimator'):
			option_value= self.parser.get('BKG_SUBTRACTION', 'bkg_estimator')	
			if option_value:
				self.bkg_estimator= option_value

		if self.parser.has_option('BKG_SUBTRACTION', 'bkg_inner_radius_factor'):
			option_value= self.parser.get('BKG_SUBTRACTION', 'bkg_inner_radius_factor')	
			if option_value:
				self.bkg_inner_radius_factor= float(option_value)

		if self.parser.has_option('BKG_SUBTRACTION', 'bkg_outer_radius_factor'):
			option_value= self.parser.get('BKG_SUBTRACTION', 'bkg_outer_radius_factor')	
			if option_value:
				self.bkg_outer_radius_factor= float(option_value)

		if self.parser.has_option('BKG_SUBTRACTION', 'bkg_max_nan_thr'):
			option_value= self.parser.get('BKG_SUBTRACTION', 'bkg_max_nan_thr')	
			if option_value:
				self.bkg_max_nan_thr= float(option_value)


		# - Parse FIRST DATA section options
		if self.parser.has_option('FIRST_DATA', 'metadata'):
			option_value= self.parser.get('FIRST_DATA', 'metadata')	
			if option_value:
				self.survey_options['first']['metadata']= option_value

		# - Parse NVSS DATA section options
		if self.parser.has_option('NVSS_DATA', 'metadata'):
			option_value= self.parser.get('NVSS_DATA', 'metadata')	
			if option_value:
				self.survey_options['nvss']['metadata']= option_value

		# - Parse MGPS DATA section options
		if self.parser.has_option('MGPS_DATA', 'metadata'):
			option_value= self.parser.get('MGPS_DATA', 'metadata')	
			if option_value:
				self.survey_options['mgps']['metadata']= option_value

		# - Parse VGPS DATA section options
		if self.parser.has_option('VGPS_DATA', 'metadata'):
			option_value= self.parser.get('VGPS_DATA', 'metadata')	
			if option_value:
				self.survey_options['vgps']['metadata']= option_value

		# - Parse SGPS DATA section options
		if self.parser.has_option('SGPS_DATA', 'metadata'):
			option_value= self.parser.get('SGPS_DATA', 'metadata')	
			if option_value:
				self.survey_options['sgps']['metadata']= option_value

		# - Parse VLASS DATA section options
		if self.parser.has_option('VLASS_DATA', 'metadata'):
			option_value= self.parser.get('VLASS_DATA', 'metadata')	
			if option_value:
				self.survey_options['vlass']['metadata']= option_value

		# - Parse WISE section options
		if self.parser.has_option('WISE_3_4_DATA', 'metadata'):
			option_value= self.parser.get('WISE_3_4_DATA', 'metadata')	
			if option_value:
				self.survey_options['wise_3_4']['metadata']= option_value

		if self.parser.has_option('WISE_4_6_DATA', 'metadata'):
			option_value= self.parser.get('WISE_4_6_DATA', 'metadata')	
			if option_value:
				self.survey_options['wise_4_6']['metadata']= option_value

		if self.parser.has_option('WISE_12_DATA', 'metadata'):
			option_value= self.parser.get('WISE_12_DATA', 'metadata')	
			if option_value:
				self.survey_options['wise_12']['metadata']= option_value

		if self.parser.has_option('WISE_22_DATA', 'metadata'):
			option_value= self.parser.get('WISE_22_DATA', 'metadata')	
			if option_value:
				self.survey_options['wise_22']['metadata']= option_value

		# - Parse HERSCHEL section options
		if self.parser.has_option('HERSCHEL_HIGAL70_DATA', 'metadata'):
			option_value= self.parser.get('HERSCHEL_HIGAL70_DATA', 'metadata')	
			if option_value:
				self.survey_options['higal_70']['metadata']= option_value

		if self.parser.has_option('HERSCHEL_HIGAL160_DATA', 'metadata'):
			option_value= self.parser.get('HERSCHEL_HIGAL160_DATA', 'metadata')	
			if option_value:
				self.survey_options['higal_160']['metadata']= option_value

		if self.parser.has_option('HERSCHEL_HIGAL250_DATA', 'metadata'):
			option_value= self.parser.get('HERSCHEL_HIGAL250_DATA', 'metadata')	
			if option_value:
				self.survey_options['higal_250']['metadata']= option_value

		if self.parser.has_option('HERSCHEL_HIGAL350_DATA', 'metadata'):
			option_value= self.parser.get('HERSCHEL_HIGAL350_DATA', 'metadata')	
			if option_value:
				self.survey_options['higal_350']['metadata']= option_value

		if self.parser.has_option('HERSCHEL_HIGAL500_DATA', 'metadata'):
			option_value= self.parser.get('HERSCHEL_HIGAL500_DATA', 'metadata')	
			if option_value:
				self.survey_options['higal_500']['metadata']= option_value

		# - Parse Spitzer IRAC section options
		if self.parser.has_option('SPITZER_IRAC3_6_DATA', 'metadata'):
			option_value= self.parser.get('SPITZER_IRAC3_6_DATA', 'metadata')	
			if option_value:
				self.survey_options['irac_3_6']['metadata']= option_value
	
		if self.parser.has_option('SPITZER_IRAC4_5_DATA', 'metadata'):
			option_value= self.parser.get('SPITZER_IRAC4_5_DATA', 'metadata')	
			if option_value:
				self.survey_options['irac_4_5']['metadata']= option_value

		if self.parser.has_option('SPITZER_IRAC5_8_DATA', 'metadata'):
			option_value= self.parser.get('SPITZER_IRAC5_8_DATA', 'metadata')	
			if option_value:
				self.survey_options['irac_5_8']['metadata']= option_value

		if self.parser.has_option('SPITZER_IRAC8_DATA', 'metadata'):
			option_value= self.parser.get('SPITZER_IRAC8_DATA', 'metadata')	
			if option_value:
				self.survey_options['irac_8']['metadata']= option_value

		# - Parse Spitzer MIPS section options
		if self.parser.has_option('SPITZER_MIPS24_DATA', 'metadata'):
			option_value= self.parser.get('SPITZER_MIPS24_DATA', 'metadata')	
			if option_value:
				self.survey_options['mips_24']['metadata']= option_value
			
		# - Parse APEX data
		if self.parser.has_option('APEX_ATLASGAL_DATA', 'metadata'):
			option_value= self.parser.get('APEX_ATLASGAL_DATA', 'metadata')	
			if option_value:
				self.survey_options['atlasgal']['metadata']= option_value

		if self.parser.has_option('APEX_ATLASGAL_PLANCK_DATA', 'metadata'):
			option_value= self.parser.get('APEX_ATLASGAL_PLANCK_DATA', 'metadata')	
			if option_value:
				self.survey_options['atlasgal_planck']['metadata']= option_value

		# - Parse MSX
		if self.parser.has_option('MSX_8_3_DATA', 'metadata'):
			option_value= self.parser.get('MSX_8_3_DATA', 'metadata')	
			if option_value:
				self.survey_options['msx_8_3']['metadata']= option_value

		if self.parser.has_option('MSX_12_1_DATA', 'metadata'):
			option_value= self.parser.get('MSX_12_1_DATA', 'metadata')	
			if option_value:
				self.survey_options['msx_12_1']['metadata']= option_value

		if self.parser.has_option('MSX_14_7_DATA', 'metadata'):
			option_value= self.parser.get('MSX_14_7_DATA', 'metadata')	
			if option_value:
				self.survey_options['msx_14_7']['metadata']= option_value

		if self.parser.has_option('MSX_21_3_DATA', 'metadata'):
			option_value= self.parser.get('MSX_21_3_DATA', 'metadata')	
			if option_value:
				self.survey_options['msx_21_3']['metadata']= option_value

		# - Parse SCORPIO ATCA section options
		if self.parser.has_option('SCORPIO_ATCA_2_1_DATA', 'metadata'):
			option_value= self.parser.get('SCORPIO_ATCA_2_1_DATA', 'metadata')	
			if option_value:
				self.survey_options['scorpio_atca_2_1']['metadata']= option_value

		# - Parse SCORPIO ASKAP section options	
		if self.parser.has_option('SCORPIO_ASKAP15_B1_DATA', 'metadata'):
			option_value= self.parser.get('SCORPIO_ASKAP15_B1_DATA', 'metadata')	
			if option_value:
				self.survey_options['scorpio_askap15_b1']['metadata']= option_value

		if self.parser.has_option('SCORPIO_ASKAP36_B123_DATA', 'metadata'):
			option_value= self.parser.get('SCORPIO_ASKAP36_B123_DATA', 'metadata')	
			if option_value:
				self.survey_options['scorpio_askap36_b123']['metadata']= option_value

		if self.parser.has_option('SCORPIO_ASKAP36_B123_CH1_DATA', 'metadata'):
			option_value= self.parser.get('SCORPIO_ASKAP36_B123_CH1_DATA', 'metadata')	
			if option_value:
				self.survey_options['scorpio_askap36_b123_ch1']['metadata']= option_value

		if self.parser.has_option('SCORPIO_ASKAP36_B123_CH2_DATA', 'metadata'):
			option_value= self.parser.get('SCORPIO_ASKAP36_B123_CH2_DATA', 'metadata')	
			if option_value:
				self.survey_options['scorpio_askap36_b123_ch2']['metadata']= option_value

		if self.parser.has_option('SCORPIO_ASKAP36_B123_CH3_DATA', 'metadata'):
			option_value= self.parser.get('SCORPIO_ASKAP36_B123_CH3_DATA', 'metadata')	
			if option_value:
				self.survey_options['scorpio_askap36_b123_ch3']['metadata']= option_value

		if self.parser.has_option('SCORPIO_ASKAP36_B123_CH4_DATA', 'metadata'):
			option_value= self.parser.get('SCORPIO_ASKAP36_B123_CH4_DATA', 'metadata')	
			if option_value:
				self.survey_options['scorpio_askap36_b123_ch4']['metadata']= option_value

		if self.parser.has_option('SCORPIO_ASKAP36_B123_CH5_DATA', 'metadata'):
			option_value= self.parser.get('SCORPIO_ASKAP36_B123_CH5_DATA', 'metadata')	
			if option_value:
				self.survey_options['scorpio_askap36_b123_ch5']['metadata']= option_value

		# - Parse THOR survey
		if self.parser.has_option('THOR_DATA', 'metadata'):
			option_value= self.parser.get('THOR_DATA', 'metadata')	
			if option_value:
				self.survey_options['thor']['metadata']= option_value

		# - Parse MeerKAT-GPS section options	
		if self.parser.has_option('MEERKAT_GPS_DATA', 'metadata'):
			option_value= self.parser.get('MEERKAT_GPS_DATA', 'metadata')	
			if option_value:
				self.survey_options['meerkat_gps']['metadata']= option_value
			
		if self.parser.has_option('MEERKAT_GPS_CH1_DATA', 'metadata'):
			option_value= self.parser.get('MEERKAT_GPS_CH1_DATA', 'metadata')	
			if option_value:
				self.survey_options['meerkat_gps_ch1']['metadata']= option_value

		if self.parser.has_option('MEERKAT_GPS_CH2_DATA', 'metadata'):
			option_value= self.parser.get('MEERKAT_GPS_CH2_DATA', 'metadata')	
			if option_value:
				self.survey_options['meerkat_gps_ch2']['metadata']= option_value

		if self.parser.has_option('MEERKAT_GPS_CH3_DATA', 'metadata'):
			option_value= self.parser.get('MEERKAT_GPS_CH3_DATA', 'metadata')	
			if option_value:
				self.survey_options['meerkat_gps_ch3']['metadata']= option_value

		if self.parser.has_option('MEERKAT_GPS_CH4_DATA', 'metadata'):
			option_value= self.parser.get('MEERKAT_GPS_CH4_DATA', 'metadata')	
			if option_value:
				self.survey_options['meerkat_gps_ch4']['metadata']= option_value

		if self.parser.has_option('MEERKAT_GPS_CH5_DATA', 'metadata'):
			option_value= self.parser.get('MEERKAT_GPS_CH5_DATA', 'metadata')	
			if option_value:
				self.survey_options['meerkat_gps_ch5']['metadata']= option_value

		if self.parser.has_option('MEERKAT_GPS_CH6_DATA', 'metadata'):
			option_value= self.parser.get('MEERKAT_GPS_CH6_DATA', 'metadata')	
			if option_value:
				self.survey_options['meerkat_gps_ch6']['metadata']= option_value

		if self.parser.has_option('MEERKAT_GPS_CH7_DATA', 'metadata'):
			option_value= self.parser.get('MEERKAT_GPS_CH7_DATA', 'metadata')	
			if option_value:
				self.survey_options['meerkat_gps_ch7']['metadata']= option_value

		if self.parser.has_option('MEERKAT_GPS_CH8_DATA', 'metadata'):
			option_value= self.parser.get('MEERKAT_GPS_CH8_DATA', 'metadata')	
			if option_value:
				self.survey_options['meerkat_gps_ch8']['metadata']= option_value

		if self.parser.has_option('MEERKAT_GPS_CH9_DATA', 'metadata'):
			option_value= self.parser.get('MEERKAT_GPS_CH9_DATA', 'metadata')	
			if option_value:
				self.survey_options['meerkat_gps_ch9']['metadata']= option_value

		if self.parser.has_option('MEERKAT_GPS_CH10_DATA', 'metadata'):
			option_value= self.parser.get('MEERKAT_GPS_CH10_DATA', 'metadata')	
			if option_value:
				self.survey_options['meerkat_gps_ch10']['metadata']= option_value

		if self.parser.has_option('MEERKAT_GPS_CH11_DATA', 'metadata'):
			option_value= self.parser.get('MEERKAT_GPS_CH11_DATA', 'metadata')	
			if option_value:
				self.survey_options['meerkat_gps_ch11']['metadata']= option_value
	
		if self.parser.has_option('MEERKAT_GPS_CH12_DATA', 'metadata'):
			option_value= self.parser.get('MEERKAT_GPS_CH12_DATA', 'metadata')	
			if option_value:
				self.survey_options['meerkat_gps_ch12']['metadata']= option_value

		if self.parser.has_option('MEERKAT_GPS_CH13_DATA', 'metadata'):
			option_value= self.parser.get('MEERKAT_GPS_CH13_DATA', 'metadata')	
			if option_value:
				self.survey_options['meerkat_gps_ch13']['metadata']= option_value

		if self.parser.has_option('MEERKAT_GPS_CH14_DATA', 'metadata'):
			option_value= self.parser.get('MEERKAT_GPS_CH14_DATA', 'metadata')	
			if option_value:
				self.survey_options['meerkat_gps_ch14']['metadata']= option_value

		# - Parse ASKAP RACS section options	
		if self.parser.has_option('ASKAP_RACS_DATA', 'metadata'):
			option_value= self.parser.get('ASKAP_RACS_DATA', 'metadata')	
			if option_value:
				self.survey_options['askap_racs']['metadata']= option_value		

		# - Parse MAGPIS 21cm section options	
		if self.parser.has_option('MAGPIS_21cm_DATA', 'metadata'):
			option_value= self.parser.get('MAGPIS_21cm_DATA', 'metadata')	
			if option_value:
				self.survey_options['magpis_21cm']['metadata']= option_value

		# - Parse CORNISH section options	
		if self.parser.has_option('CORNISH_DATA', 'metadata'):
			option_value= self.parser.get('CORNISH_DATA', 'metadata')	
			if option_value:
				self.survey_options['cornish']['metadata']= option_value

		# - Parse custom survey section options	
		if self.parser.has_option('CUSTOM_SURVEY_DATA', 'metadata'):
			option_value= self.parser.get('CUSTOM_SURVEY_DATA', 'metadata')	
			if option_value:
				self.survey_options['custom_survey']['metadata']= option_value
	

	def parse(self, filename, add_survey=False, metadata_path=""):
		""" Read input INI config file and set options """

		# ***************************
		# **    READ CONFIG FILE
		# ***************************
		# - Read config parser
		if self.__read(filename)<0:
			logger.error("Failed to read config file %s!" % (filename))
			return -1

		# ***************************
		# **    ADD CUSTOM SURVEY
		# ***************************
		if add_survey:
			if self.__add_custom_survey(metadata_path)<0:
				logger.error("Failed to add custom survey with metadata %s!" % (metadata_path))
				return -1

		# ***************************
		# **  PARSE AND SET OPTIONS
		# ***************************
		self.__set_options()

		# ***************************
		# **    VALIDATE CONFIG
		# ***************************
		if self.__validate()<0:
			logger.error("Invalid configuration options detected, please check!")
			return -1

		return 0

	#==============================
	#     VALIDATE CONFIG
	#==============================
	def __validate(self):
		""" Validate config file """
		
		# - Check survey options
		valid_surveys= list(self.survey_options.keys())
		logger.info("Checking that search surveys exist in supported survey list ...")
		#print(valid_surveys)
		for survey in self.surveys:
			if not survey in valid_surveys:
				logger.error("Survey " + survey + " is unknown and/or not supported!")
				return -1

		# - Check survey metadata is not empty and existing file
		for survey in self.surveys:
			metadata= self.survey_options[survey]['metadata']
			if not metadata:
				logger.error("Metadata path for survey " + survey + " is empty string (please specify file in config)!")
				return -1
			if not os.path.isfile(metadata):
				logger.error("Metadata file " + metadata + " is not existing in filesystem!")
				return -1

		# ...
		
		return 0

	#==============================
	#     PRINT OPTIONS
	#==============================
	def __print_options(self):
		""" """
	
		sections= self.parser.sections()
		for section in sections:
			print("section=%s" % section)
			options= self.parser.items(section)
			print(options)
			
			for option in options:
				print(option)
				print(type(option))
			#for key, value in options.iteritems():
			#for key, value in section.iteritems():
				#print("Option %s=%s" % (key,str(value)))
		

